import numpy as np


def quantize_frequencies(x, rate=4.096e9, n_samples=2 ** 19):
    x = np.asarray(x) if not isinstance(x, (int, float)) else x
    freq_res = rate / n_samples
    return np.round(x / freq_res) * freq_res


def opfb_bin_spectrum(data, bins=None, norm_max=True, shift=True, left_snip=1, db=True):
    """
    Inputs:
    - bin: OPFB bin number (allowed values 0 - 4095)
    - data: Raw data out of OPFB. Should be in the form N x 4096 where N is the number of samples from a single bin.
    - norm_max: boolean. Normalize the spectrum to the max value (easier to read spur, etc. height)
    - shift: boolean. Either apply to every OPFB bin or don't (this should always be applied).
    - left_snip: Cuts out the left most sample in a bin. This is an annoying crutch to handle the
        fact we place the N/2 bin on the far left.
        See https://www.gaussianwaves.com/2015/11/interpreting-fft-results-complex-dft-frequency-bins-and-fftshift/
        for explanation.

    Output:
    - Returns spectrum of a single FFT bin.
    """
    bins = np.asarray(bins).astype(int)
    if db:
        x = 20 * np.log10(np.abs(np.fft.fft(data, axis=0)))
    if shift:
        x = np.fft.fftshift(x, axes=0)
    if norm_max:
        x -= np.max(x)
    return np.flip(x[left_snip:, bins], axis=0)


"""
FFT bin ordering is confusing.

We find it more convenient to work with bins in order of increasing central frequency and call this
default or native order. Though "natural" may be tempting it is overloaded as the bit order of the FFT 
algorithm is often referred to as either "bit-reversed" or "natural" order for word outputs. Our FFT outputs 
words in the natural order. 

The SSR FFT used in the OPFB produces bins in a shifted order: 
    [-1, 1), ..., [2046, 2048), [2047, -2047), ..., [-2, -1)
this on-fpga order that is ingested by bin_to_res we refer to as the raw ssr order.
"""

def opfb_bin_number(freq, ssr_raw_order=True):
    """
    Compute the OPFB bin number corresponding to each frequency, specified in Hz.

    Frequencies are assumed to be in [-2.048, 2.048) GHz. Frequencies are placed in
    the closest bin by central frequency. Bins are 2 MHz wide and centered on 1 MHz
    increments (50% overlapping).

    bins are indexed 0 to 4095 from left to right across the 4 GHz IQ spectrum.
    bin 0 is centered at -2048 MHz and contains [2047, -2047) MHz.
    bin 1 contains [-2048, -2046) MHz.
    bin 2048 contains [-1, 1) MHz.
    bin 4095 contains [2046, 2048) MHz.

    ssr_raw_order: bool
        If true return the bin number in the order of the OPFB output.
        If false return
        the OPFB bins are in order received from the Xilinx SSR FFT with no shift applied
        bin 0 contains frequencies in the interval [-1, 1) MHz, bin 1 [0, 2) MHz...
        bin 2047 contains [2046, 2048) MHz.
        The nyquist frequency component in bin 2048 is common to both positive
        and negative frequencies. By convention, it is the highest negative frequency so
        bin 2048 contains the highest negative frequency [2047, -2047) MHz
        bins 2049 + contain negative frequencies progressing towards 0 MHz.

    """
    if ssr_raw_order:
        return (np.round(freq / 1e6).astype(int) + 4096) % 4096
    else:
        return np.round(freq / 1e6).astype(int) + 2048


def opfb_bin_frequencies(bins, resolution, Fs=4.096e9, M=4096, OS=2, left_snip=1):
    """
    Inputs:
    - bin: OPFB bin (0-4095). bin 0 contains -2048 to -2046 MHz, bin 4095 contains 2045 to 2047 MHz.
    - resolution: the number of samples from a single bin to take the FFT of. This dictates the frequency
        resolution in a single OPFB bin.
    - Fs: ADC Sampling Rate.
    - M: OPFB FFT Size.
    - OS: Oversample ratio.
    - left_snip: Cuts out the left most sample in a bin. This is an annoying crutch to handle the
        fact we place the N/2 bin on the far left.
        See https://www.gaussianwaves.com/2015/11/interpreting-fft-results-complex-dft-frequency-bins-and-fftshift/
        for explanation.

    Outputs:
    - Returns an array of frequency values in Hz for the given OPFB bin. """
    try:
        len(bins)
    except TypeError:
        bins = [bins]
    bins = np.asarray(bins).astype(int)
    bin_centers = (Fs / M) * np.linspace(-M / 2, M / 2 - 1, M)
    bin_width = (Fs / M) * OS
    base_freq = np.linspace(-bin_width / 2, bin_width / 2 - 1, resolution - left_snip)
    return base_freq[:, None] + bin_centers[bins]


def opfb_bin_center(bins, Fs=4.096e9, M=4096, ssr_order=True):
    try:
        len(bins)
    except TypeError:
        bins = [bins]
    bins = np.asarray(bins).astype(int)
    f=(Fs / M) * np.linspace(-M / 2, M / 2 - 1, M)
    return np.fft.fftshift(f)[bins] if ssr_order else f[bins]


def do_fixed_point_pfb(fpcomb, fpcoeff, n_convert=None, truncate=True):
    """Set truncate to false to preserve the full output bitwidth. Truncation is done with FpBinary defaults."""
    n_total_packets = fpcomb.size // 2048 // 2 - 16 if n_convert is None else n_convert
    fft_block = np.zeros((n_total_packets+1, 256, 16), dtype=np.complex64)
    for i in range(0, n_total_packets, 2):  # each packet of ADC samples, 128 new things to a lane 2 packets to feed all channels
        lane_out = np.zeros((2, 256, 16), dtype=np.complex64)
        for l in range(16):
            fresh = np.array([fpcoeff[l, :, 7 - c_i] * fpcomb[i + 2 * c_i:i + 2 * c_i + 2, l::16, :].reshape(256, 2).T
                              for c_i in range(8)]).sum(axis=0)
            delay = np.roll(np.array(
                [fpcoeff[l, :, 7 - c_i] * fpcomb[1 + i + 2 * c_i:1 + i + 2 * c_i + 2, l::16, :].reshape(256, 2).T
                 for c_i in range(8)]).sum(axis=0), 128, axis=1)
            # Sum the multiplies are roll the delayed samples
            if truncate:
                outformat = (-9, sum(fpcomb.flat[0].format) + 9)
                conv = lambda a: np.array(list(map(lambda x: float(x.resize(outformat)), a)))
                lane_out[0, :, l] = conv(fresh[0]) + conv(fresh[1]) * 1j
                lane_out[1, :, l] = conv(delay[0]) + conv(delay[1]) * 1j
            else:
                lane_out[0, :, l] = fresh[0].astype(float) + fresh[1].astype(float) * 1j
                lane_out[1, :, l] = delay[0].astype(float) + delay[1].astype(float) * 1j
        fft_block[i] = lane_out[0]
        fft_block[i + 1] = lane_out[1]
    return fft_block



import time
import copy

import numpy as np
import numpy.typing as nt

from typing import Optional, Type, NewType
from types import NoneType
from dataclasses import dataclass
from abc import ABC

from mkidgen3.server.feedline_config import WaveformConfig

LO_QUANT = 0.596

# TODO: Common types with powersweep_helpers
InputAtten = NewType("InputAtten", float)
OutputAtten = NewType("OutputAtten", float)

# Power Gain
Gain = NewType("Gain", float)
# Total system gain normalized to the total system gain with
# both adc and dac attens set to 0
AttenRefered = NewType("AttenRefered", Gain)


def attens_to_refered(attens: tuple[OutputAtten, InputAtten]) -> AttenRefered:
    return -attens[0] - attens[1]


# Assumes data is field quantity
def apply_gain(data: nt.NDArray[np.float64 | np.complex64], gain: Type[Gain]):
    return data * np.sqrt(10 ** (gain / 10))


@dataclass
class SweepConfig:
    steps: nt.NDArray[np.float64]
    waveform: WaveformConfig
    lo_center: float | np.float64 = 6000.0
    average: int = 1024
    attens: Optional[tuple[OutputAtten, InputAtten]] = None
    tap: str = "ddciq"
    rmses: bool = True

    _idle = 0.0
    _flush = 0

    @classmethod
    def from_bandwidth(
        cls,
        bandwidth: float | np.float64,
        points: int,
        waveform: WaveformConfig,
        lo_center: float | np.float64 = 6000.0,
        average: int = 1024,
        attens: Optional[tuple[OutputAtten, InputAtten]] = None,
    ) -> "SweepConfig":
        spacing = LO_QUANT * np.floor(((bandwidth / points) / LO_QUANT))

        # These are not going to be actually quantized properly because FP but the LO algo on the IF Board is FP too.
        # We will probably want to fix both of these at some point in the near future as the IF Board is FP32
        steps = (np.arange(points) * spacing - spacing * points // 2) / 1e6

        return SweepConfig(steps, waveform, lo_center, average, attens)

    def reset_ddc(self, ddccontrol) -> "SweepConfig":
        ddccontrol.reset_phase_center()
        return self

    def run_sweep(self, ifboard, capture, progress = True) -> "Sweep":
        tones = self.waveform.waveform.quant_freqs
        iq = np.empty((tones.size, self.steps.size), np.complex64)
        if self.rmses:
            rms = np.empty((tones.size, self.steps.size), np.complex64)
        else:
            rms = None
        if self.attens:
            ifboard.set_attens(
                input_attens=self.attens[1], output_attens=self.attens[0]
            )

        it = enumerate(self.steps)
        if progress:
            import tqdm.notebook as tqdm

        it = tqdm.tqdm(it, total=len(self.steps), desc="FREQ")
        for i, step in it:
            ifboard.set_lo(step + self.lo_center)
            piq, prms = self._get_iq_point_rms(capture)
            iq[::, i] = piq[: tones.size]
            if self.rmses:
                rms[::, i] = prms[: tones.size]
        ifboard.set_lo(self.lo_center)
        return Sweep(iq, (rms / np.sqrt(self.average)) if self.rmses else rms, self)

    def frequencies(self) -> nt.NDArray[np.float64]:
        tones = self.waveform.waveform.quant_freqs
        freqs = np.zeros((tones.size, self.steps.size), dtype=np.float64)
        freqs += self.lo_center * 1e6
        freqs += self.steps * 1e6
        freqs += tones.reshape((tones.size, 1))
        return freqs

    def _get_iq_point_rms(self, capture):
        """
        Args:
            n: int
            how many points to average
        Returns: a single averaged iq data point captured from res channel 0
        """
        tones = self.waveform.waveform.quant_freqs
        if self._idle:
            time.sleep(self._idle)
        if self._flush:
            x = capture.capture_iq(self._flush, tap_location=self.tap)
            del x

        # CortexA53 is not good at math... try and keep to f32
        x = capture.capture_iq(self.average, tap_location=self.tap)
        ps_buf = np.empty((self.average, tones.size, 2), dtype=np.float32)
        ps_buf[::, ::, ::] = np.ascontiguousarray(
            x[::, : tones.size, ::], dtype=np.int16
        )
        if hasattr(self, "verify"):
            ps_buf_old = np.copy(ps_buf)
        del x

        means = np.sum(ps_buf, axis=0) / np.float32(self.average)
        if self.rmses:
            ps_buf -= means.reshape((1, tones.size, 2))
            ps_buf *= ps_buf
            rmses = np.sqrt(np.sum(ps_buf, axis=0) / np.float32(self.average))

        if hasattr(self, "verify"):
            means_golden = np.empty((tones.size, 2), dtype=np.float64)
            ps_buf_old.mean(axis=0, out=means_golden)
            assert np.allclose(means_golden, means)
            if rmses:
                rmses_golden = np.empty((tones.size, 2), dtype=np.float64)
                ps_buf_old.std(axis=0, out=rmses_golden)
                assert np.allclose(rmses_golden, rmses)

        if self.rmses:
            return means[::, 0] + 1j * means[::, 1], rmses[::, 0] + 1j * rmses[::, 1]
        return means[::, 0] + 1j * means[::, 1], None


@dataclass
class SmoothSweepConfig(SweepConfig):
    def run_sweep(self, ifboard, capture, progress = True, cache=None) -> "SmoothSweep":
        tones = self.waveform.waveform.quant_freqs
        iq = np.empty((tones.size, self.steps.size), np.complex64)
        if self.rmses:
            rms = np.empty((tones.size, self.steps.size), np.complex64)
        else:
            rms = None
        if self.attens:
            ifboard.set_attens(
                input_attens=self.attens[1], output_attens=self.attens[0]
            )

        if cache is None:
            cache = []
        if cache == []:
            for step in self.steps:
                ifboard.set_lo(self.lo_center + step)
                cache.append(ifboard.trf_control.get_certificate())

        it = enumerate(self.steps)
        if progress:
            import tqdm.notebook as tqdm

        it = tqdm.tqdm(it, total=len(self.steps), desc="FREQ")
        for i, step in it:
            ifboard.set_lo(cache[i])
            piq, prms = self._get_iq_point_rms(capture)
            iq[::, i] = piq[: tones.size]
            if self.rmses:
                rms[::, i] = prms[: tones.size]
        ifboard.set_lo(self.lo_center)
        return SmoothSweep(iq, (rms / np.sqrt(self.average)) if self.rmses else rms, self, certificates=cache)


@dataclass
class CombSweepConfig(SweepConfig):
    overlap: int = 1

    def __post_init__(self):
        tones = self.waveform.waveform.quant_freqs
        if not np.allclose(
            tones, tones[0] + np.arange(tones.size) * (tones[1] - tones[0])
        ):
            raise ValueError(
                "Waveform required to have evenly spaced comb in incresing frequency order"
            )

    @classmethod
    def from_comb(
        cls,
        points: int,
        overlap: int,
        waveform: WaveformConfig,
        lo_center: float | np.float64 = 6000.0,
        average: int = 1024,
        attens: Optional[tuple[OutputAtten, InputAtten]] = None,
    ) -> "SweepConfig":
        if overlap > points:
            raise ValueError("overlap must be less than points")
        if overlap < 1:
            raise ValueError("Overlap should be >= 1 to enable stitching")

        tones = waveform.waveform.quant_freqs

        steps = np.linspace(0, tones[1] - tones[0], points, endpoint=False) / 1e6
        steps_overlap = steps[:overlap] + (tones[1] - tones[0]) / 1e6
        steps = np.concatenate((steps, steps_overlap))

        return CombSweepConfig(
            steps, waveform, lo_center, average, attens, overlap=overlap
        )

    def run_sweep(self, ifboard, capture, progress=True) -> "CombSweep":
        s = super().run_sweep(ifboard, capture, progress)
        return CombSweep(s.iq, s.iqsigma, self)


@dataclass
class AbstractSweep(ABC):
    iq: nt.NDArray[np.complex64]
    iqsigma: nt.NDArray[np.complex64]


@dataclass
class Sweep(AbstractSweep):
    iq: nt.NDArray[np.complex64]
    iqsigma: Optional[nt.NDArray[np.complex64]]
    config: "SweepConfig"
    equalized_gain: Optional[AttenRefered] = None

    @property
    def frequencies(self) -> nt.NDArray[np.float64]:
        return self.config.frequencies()

    def plot(
        self,
        ax,
        stacked: bool = False,
        newtones: Optional[list[float] | nt.NDArray[np.float64]] = None,
        channels: Optional[slice] = None,
        label_tones: bool = False,
        power: bool = True,
        **kwargs,
    ):
        if channels is None:
            channels = slice(0, self.iq.shape[0])
        for i in range(self.iq[channels, :].shape[0]):
            line = ax.plot(
                (self.frequencies[channels][i] / 1e6 if not stacked else self.config.steps),
                np.abs(self.iq[channels][i]) if not power else 20*np.log10(np.abs(self.iq[channels][i])),
                label=(
                    "Tone: {:.3f} MHz".format(
                        self.config.waveform.default_ddc_config.tones[channels][i] / 1e6
                    )
                    if label_tones
                    else None
                ),
                **kwargs,
            )
            if newtones:
                if stacked:
                    ax.axvline(
                        newtones[channels][i] / 1e6
                        - self.config.waveform.default_ddc_config.tones[channels][i] / 1e6,
                        color=line[0].get_color(),
                        linestyle="--",
                    )
                else:
                    ax.axvline(
                        newtones[channels][i] / 1e6 + self.config.lo_center,
                        color=line[0].get_color(),
                        linestyle="--",
                    )
        if stacked:
            ax.axvline(0, color="black", lw=0.1, label="LO")
        if label_tones:
            ax.legend()
        ax.set_ylabel("S21 (Magnitude)" if not power else "S21 (Power [dB])")
        ax.set_xlabel("Frequency (MHz)")

    def plot_loops(self, ax, channels: Optional[slice] = None, newtones: Optional[list[float] | nt.NDArray[np.float64]] = None, **kwargs):
        if channels is None:
            channels = slice(0, self.iq.shape[0])
        for i in range(self.iq[channels, :].shape[0]):
            ax.plot(self.iq[channels][i].real, self.iq[channels][i].imag, "-o", **kwargs)
        if newtones is not None:
            newtones = np.array(newtones)
            freqs = self.frequencies - self.config.lo_center * 1e6
            for i, t in enumerate(newtones[channels]):
                v = np.argmin(np.abs(freqs[channels][i] - t))
                ax.scatter(self.iq[channels][i][v].real, self.iq[channels][i][v].imag, marker="*", c='r', zorder=len(newtones) + 1, s = 24)
        ax.set_aspect("equal")
        ax.set_xlabel("I")
        ax.set_ylabel("Q")

    def equalize(self, refered_gain: AttenRefered):
        if self.equalized_gain is not None:
            current_gain = self.equalized_gain
        elif self.config.attens is not None:
            current_gain = attens_to_refered(self.config.attens)
        else:
            raise ValueError(
                "Neither the current equalized gain, nor the original gain is known"
            )
        applied_gain = refered_gain - current_gain
        return self.__class__(
            iq=apply_gain(self.iq, applied_gain),
            iqsigma=apply_gain(self.iqsigma, applied_gain),
            config=self.config,
            equalized_gain=refered_gain,
        )
@dataclass
class SmoothSweep(Sweep):
    certificates: Optional[list["TRFCalibrationCertificate"]] = None

@dataclass
class CombSweep(Sweep):
    config: "CombSweepConfig"

    def stitch(self) -> "StitchedSweep":
        freqs = self.frequencies
        iq = self.iq
        iqs = self.iqsigma

        iq_new = np.empty(
            (
                1,
                (iq.shape[1] - self.config.overlap) * iq.shape[0] + self.config.overlap,
            ),
            np.complex64,
        )
        iqs_new = np.empty_like(iq_new)
        freqs_new = np.empty_like(iq_new, np.float64)

        iq_new[0][: iq.shape[1]] = iq[0]
        iqs_new[0][: iq.shape[1]] = iqs[0]
        freqs_new[0][: iq.shape[1]] = freqs[0]
        for i in range(1, self.frequencies.shape[0]):
            low = slice(
                (iq.shape[1] - self.config.overlap) * i,
                (iq.shape[1] - self.config.overlap) * i + self.config.overlap,
            )
            high = slice(self.config.overlap)
            high_new = slice(
                (iq.shape[1] - self.config.overlap) * i + self.config.overlap,
                (iq.shape[1] - self.config.overlap) * (i + 1) + self.config.overlap,
            )
            assert np.all(freqs_new[0][low] == freqs[i][high])
            ratio = iq_new[0][low] / iq[i][high]
            print(np.mean(ratio), np.abs(np.mean(ratio)))
            ratiosigma = (
                np.sqrt(
                    (iqs_new[0][low] / iq_new[0][low]) ** 2
                    + (iqs[i][high] / iq[i][high]) ** 2
                )
                * ratio
            )
            ratio, ratiosigma = np.sum(ratio / (ratiosigma**2)) / np.sum(
                1 / (ratiosigma**2)
            ), 1 / np.sqrt(np.sum(1 / (ratiosigma**2)))
            iq_new[0][high_new] = iq[i][self.config.overlap :] * ratio
            iqs_new[0][high_new] = iq_new[0][high_new] * np.sqrt(
                (iqs[i][self.config.overlap :] / iq[i][self.config.overlap :]) ** 2
                + (ratiosigma / ratio) ** 2
            )
            freqs_new[0][high_new] = freqs[i][self.config.overlap :]
        return StitchedSweep(
            iq=iq_new,
            iqsigma=iqs_new,
            original=self,
            equalized_gain=self.equalized_gain,
            frequencies=freqs_new,
        )


@dataclass
class StitchedSweep(AbstractSweep):
    original: "CombSweep"
    frequencies: nt.NDArray[np.float64]
    equalized_gain: Optional[AttenRefered] = None

    def equalize(self, refered_gain: AttenRefered):
        if self.equalized_gain is not None:
            current_gain = self.equalized_gain
        elif self.original.config.attens is not None:
            current_gain = attens_to_refered(self.original.config.attens)
        else:
            raise ValueError(
                "Neither the current equalized gain, nor the original gain is known"
            )
        applied_gain = current_gain - refered_gain
        return self.__class__(
            iq=apply_gain(self.iq, applied_gain),
            iqsigma=apply_gain(self.iqsigma, applied_gain),
            original=self.original,
            equalized_gain=refered_gain,
            frequencies=self.frequencies,
        )


@dataclass
class PowerSweepConfig:
    attens: dict[OutputAtten, InputAtten]
    sweep_config: Type[SweepConfig]

    def __post_init__(self):
        # TODO: Check attens valid
        if self.sweep_config.attens is not None:
            raise ValueError(
                "Input and Output attens must be unset in the template config"
            )

    @classmethod
    def from_matched(
        cls,
        starting_output: OutputAtten,
        starting_input: InputAtten,
        output_step: OutputAtten,
        steps: int,
        sweep_config: Type[SweepConfig],
    ):
        attens = {}
        for i in range(steps):
            attens[starting_output + output_step * i] = min(
                max(starting_input - output_step * i, 0), 31.75 * 2
            )
        return PowerSweepConfig(attens, sweep_config)

    def run_powersweep(self, ifboard, capture, progress=False) -> "PowerSweep":
        sweeps = {}
        iter = self.attens.items()
        kwargs = {}
        if isinstance(self.sweep_config, SmoothSweepConfig):
            kwargs["cache"] = []
        if progress:
            import tqdm.notebook as tqdm

            iter = tqdm.tqdm(iter, total=len(list(self.attens.keys())), desc="ATTN")
        for output_atten, input_atten in iter:
            this_sweepconfig = copy.copy(self.sweep_config)
            this_sweepconfig.attens = (output_atten, input_atten)
            sweeps[output_atten] = (
                input_atten,
                this_sweepconfig.run_sweep(ifboard, capture, progress, **kwargs),
            )
        return PowerSweep(self, sweeps)


@dataclass
class PowerSweep:
    config: "PowerSweepConfig"
    sweeps: dict[OutputAtten, tuple[InputAtten, Type[Sweep]]]

    def plot(
        self,
        fig,
        ax,
        output_range: Optional[tuple[OutputAtten, OutputAtten]] = None,
        equalize: Optional[AttenRefered] = 0,
        power: bool = True,
        channels: Optional[slice] = None,
        cmap=None,
    ):
        if cmap is None:
            import matplotlib.pyplot as plt

            cmap = plt.cm.viridis_r
        if output_range is not None and output_range[0] > output_range[1]:
            output_range = (output_range[1], output_range[0])
        sweeps = (
            self.sweeps
            if output_range is None
            else {
                o: (i, s)
                for o, (i, s) in self.sweeps.items()
                if o >= output_range[0] and o <= output_range[1]
            }
        )
        sweeps = (
            sweeps
            if equalize is None
            else {o: (i, s.equalize(equalize)) for o, (i, s) in sweeps.items()}
        )
        output_min = (
            min(list(sweeps.keys())) if output_range is None else output_range[0]
        )
        output_max = (
            max(list(sweeps.keys())) if output_range is None else output_range[1]
        )
        import matplotlib.colors as cl
        import matplotlib.cm as cm

        norm = cl.Normalize(output_min, output_max)
        smap = cm.ScalarMappable(norm, cmap)
        fig.colorbar(smap, ax=ax, label="Output Attenuation (dB)")
        for output_atten, (_, sweep) in sweeps.items():
            sweep.plot(ax, channels = channels, color=cmap(norm(output_atten)))



import zmq
from mkidgen3.server.feedline_config import IFConfig, BitstreamConfig, RFDCClockingConfig, RFDCConfig, WaveformConfig, ChannelConfig, DDCConfig, FeedlineConfig, FilterConfig, TriggerConfig
from mkidgen3.server.captures import CaptureJob, FRSClient, CaptureRequest, StatusListener
from mkidgen3.server.waveform import WaveformFactory
from mkidgen3.opfb import opfb_bin_number
import numpy as np
from mkidgen3.util import setup_logging
setup_logging('feedlineclient')
# ctx = zmq.Context.instance()
# ctx.linger = 0

frs = 'a'

if frs == 'a':
    server = FRSClient(url='mkidrfsoc4x2.physics.ucsb.edu', command_port=8888, data_port=8889, status_port=8890)
elif frs == 'b':
    server = FRSClient(url='rfsoc4x2b.physics.ucsb.edu', command_port=8888, data_port=8889, status_port=8890)
# Status listener
#gsm = StatusListener(b'', server.status_url)

# Bitstream Config
bitstream = BitstreamConfig(bitstream='/home/xilinx/gen3_top_final.bit', ignore_version=True)
# RFDC Clocking Config
rfdc_clk = RFDCClockingConfig(programming_key='4.096GSPS_MTS_dualloop', clock_source=None)  # clock source should default to external 10 MHz
# RDFC Config
rfdc = RFDCConfig(dac_mts=True, adc_mts=False, adc_gains=None, dac_gains=None)
# IF Config
if_board0 = IFConfig(lo=6000, adc_attn=50, dac_attn=50)
if_board1 = IFConfig(lo=7000, adc_attn=40, dac_attn=30)
if_cfgs = [if_board0, if_board1]
# DAC Config
waveform0 = WaveformConfig(waveform=WaveformFactory(n_uniform_tones=512))
waveform1 = WaveformConfig(waveform=WaveformFactory(n_uniform_tones=1024))
output_waveform = waveform0.waveform.output_waveform # trigger waveform computation
#output_waveform = waveform1.waveform.output_waveform # trigger waveform computation
wvfm_cfgs = [waveform0, waveform1]
# Bin2Res Config
chan0 = waveform0.default_channel_config
chan1 = waveform1.default_channel_config
b2r_cfgs = [chan0, chan1]
# DDC Config
ddc0 = waveform0.default_ddc_config
ddc1 = waveform1.default_ddc_config
ddc_cfgs = [ddc0, ddc1]
# Feedline Config
fcs = []
for i in [0, 1]:
    fcs.append(FeedlineConfig(bitstream=bitstream, rfdc_clk=rfdc_clk, rfdc=rfdc,
                    if_board=if_cfgs[i], waveform=wvfm_cfgs[i], chan=b2r_cfgs[i], ddc=ddc_cfgs[i]))

cr = CaptureRequest(1024, 'ddciq', fcs[0], server, channels=[4, 8, 12])
j = CaptureJob(cr)
j.submit(True, True)


crs = []
crs.append(CaptureRequest(2**14, 'adc', fcs[0], server))
crs.append(CaptureRequest(2**29, 'adc', fcs[0], server))
crs.append(CaptureRequest(2**7, 'adc', fcs[1], server))
crs.append(CaptureRequest(2**14, 'ddciq', fcs[0], server))
crs.append(CaptureRequest(2, 'ddciq', fcs[0], server))
crs.append(CaptureRequest(2, 'ddciq', fcs[0], server, channels=[0, 2, 3]))
crs.append(CaptureRequest(2**16, 'ddciq', fcs[1], server, channels=[0, 2, 3]))
crs.append(CaptureRequest(2**14, 'filtphase', fcs[0], server, channels=[0, 1, 2, 3, 4, 5, 6, 7, 8]))
crs.append(CaptureRequest(2**8, 'filtphase', fcs[1], server))
crs.append(CaptureRequest(2**26, 'filtphase', fcs[0], server, channels=[8, 9, 10]))

jobs = []
for i, cr in enumerate(crs):
    jobs.append(CaptureJob(cr))
    jobs[i].submit(True, True)

print('hi')
print('hi')



import pynq
from pynq import DefaultHierarchy
from logging import getLogger

from mkidgen3.drivers.rfdcclock import configure
from mkidgen3.mkidpynq import get_board_name


def status():
    import mkidgen3
    rfdc = mkidgen3._gen3_overlay.usp_rf_data_converter_0

    regmap = {'Restart Power-On State Machine': 0x0004,
              'Restart State': 0x0008,
              'Current State': 0x000C,
              'Reset Count': 0x0038,
              'Interrupt Status': 0x0200,
              'Tile Common Status': 0x0228,
              'Tile Disable': 0x0230}
    tilemap = [(f'ADC{i}', v) for i, v in enumerate((0x14000, 0x18000, 0x1C000, 0x20000))]
    tilemap += [(f'DAC{i}', v) for i, v in enumerate((0x04000, 0x08000))]  # , 0x0C000, 0x10000))]
    tilemap = dict(tilemap)
    print(rfdc.read(0x0008))
    for t, taddr in tilemap.items():
        print(t)
        for k, r in regmap.items():
            print(f'  {k}:  {rfdc.read(taddr + r)}')


def reset():
    import mkidgen3
    rfdc = mkidgen3._gen3_overlay.usp_rf_data_converter_0
    rfdc.write(0x0004, 0x00000001)


# This does not appear to actually be necessary for MTS but the example does do it.
def reset_clockwizard(overlay:pynq.Overlay):
    CLOCKWIZARD_LOCK_ADDRESS = 0x0004
    CLOCKWIZARD_RESET_ADDRESS = 0x0000
    CLOCKWIZARD_RESET_TOKEN = 0x000A
    mmcm = overlay.Clocktree.RF_CLKGEN
    mmcm.mmio.write_reg(CLOCKWIZARD_RESET_ADDRESS, CLOCKWIZARD_RESET_TOKEN)


class RFDCHierarchy(DefaultHierarchy):
    def __init__(self, description):
        super().__init__(description)
        self.rfdc = self.usp_rf_data_converter_0
        try:
            self.switch = self.axis_switch_0
        except AttributeError:
            self.switch = None
            getLogger(__name__).info('RFDCHierarchy does not support switching ADCs')

    def start_clocks(self, external_10mhz=False):
        configure(external_10mhz)

    def reset(self):
        self.rfdc.write(0x0004, 0x00000001)

    def select_adc(self, adc='single_ended'):
        if self.switch is None:
            raise RuntimeError('RFDCHierarchy does not support switching ADCs')
        self.switch.set_driver(slave=int(adc == 'single_ended'))

    @property
    def active_adc(self):
        if self.switch is None:
            return 'Switching not supported. Please add driver introspection to determine active adc'
        if self.switch.is_disabled():
            return 'none'
        elif self.switch.driver_for() & 1:
            return 'single_ended (ADC0 0,2)'
        else:
            return 'differential (ADC2 0,2)'

    def rfdc_status(self, tell=False):
        regmap = {'Restart Power-On State Machine': 0x0004,
                  'Restart State': 0x0008,
                  'Current State': 0x000C,
                  'Reset Count': 0x0038,
                  'Interrupt Status': 0x0200,
                  'Tile Common Status': 0x0228,
                  'Tile Disable': 0x0230}
        tilemap = [(f'ADC{i}', v) for i, v in enumerate((0x14000, 0x18000, 0x1C000, 0x20000))]
        tilemap += [(f'DAC{i}', v) for i, v in enumerate((0x04000, 0x08000))]  # , 0x0C000, 0x10000))]
        tilemap = dict(tilemap)

        ret = {'global': self.rfdc.read(0x0008)}
        for t, taddr in tilemap.items():
            ret[t] = {k: self.rfdc.read(taddr + r) for k, r in regmap.items()}
        if tell:
            print('Global Status: ' + hex(ret['global']))
            for t in ([f'ADC{i}' for i in range(4)] + [f'DAC{i}' for i in range(2)]):
                print(f'{t}:')
                for k, v in ret[t].items():
                    print(f'  {k}: {v}')
        return ret

    @staticmethod
    def checkhierarchy(description):
        if 'usp_rf_data_converter_0' not in description['ip']:
            return False
        return True

    def enable_mts(self, dac=True, adc=False):
        """Synchronizes all active ADC and DAC tiles

        Parameters
        ----------
        dac : Boolean
            Enables MTS on the DAC
        adc : Boolean
            Enables MTS on the ADC
        """
        if not (dac or adc):
            return

        if get_board_name() == "RFSoC4x2":
            self.ACTIVE_DAC_TILES = 0b0101
            self.ACTIVE_ADC_TILES = 0b0101
            self.MAX_DAC_TILES = 4
            self.MAX_ADC_TILES = 4
            self.DAC_REF_TILE = 2
            self.ADC_REF_TILE = 2
        else:
            raise NotImplementedError(f"{get_board_name()} MTS Not Supported")

        if dac:
            self.rfdc.mts_dac_config.RefTile = self.DAC_REF_TILE
        if adc:
            self.rfdc.mts_adc_config.RefTile = self.ADC_REF_TILE

        self.init_tile_sync(dac=dac, adc=adc)
        self.sync_tiles(dac=dac, adc=adc)

    def init_tile_sync(self, reset_clockwiz=False, dac=True, adc=True):
        """Initilizes the ADCs and DACs for MTS

        This resets all the DACs and ADCs, initilizes the MTS engine in the tiles with the CLK
        inputs and turns the rest of the tiles back on

        Parameters
        ----------
        reset_clockwiz : Boolean
            Resets the clockwizard driving the design, this does not appear to be required
        dac : Boolean
            Enables MTS on the DAC
        adc : Boolean
            Enables MTS on the ADC
        """
        import time

        # It's gross but I'm not going to mess with the structure here because there is some
        # weirdness around the dac config some of which seems to come from the ADC. Unfortunately
        # none of the MTS regs are documented and I haven't had the time to debug said weirdness
        if dac:
            self.rfdc.mts_dac_config.Tiles = 0b0001  # turn only one tile on first
        if adc:
            self.rfdc.mts_adc_config.Tiles = 0b0001
        # Start the SYSREF capture engine in the first tiles
        if dac:
            self.rfdc.mts_dac_config.SysRef_Enable = 1
            self.rfdc.mts_dac_config.Target_Latency = -1
        if adc:
            self.rfdc.mts_adc_config.SysRef_Enable = 1
            self.rfdc.mts_adc_config.Target_Latency = -1
        #
        if dac:
            self.rfdc.mts_dac()
        if adc:
            self.rfdc.mts_adc()
        # Reset MTS ClockWizard MMCM - refer to PG065
        if reset_clockwiz:
            reset_clockwizard(self._overlay)
        time.sleep(0.1)
        # Reset only user selected DAC tiles
        if dac:
            bitvector = self.ACTIVE_DAC_TILES
            for n in range(self.MAX_DAC_TILES):
                if (bitvector & 0x1):
                    self.rfdc.dac_tiles[n].Reset()
                bitvector = bitvector >> 1
        # Reset ADC FIFO of only user selected tiles - restarts MTS engine
        if adc:
            for toggleValue in range(0, 1):
                bitvector = self.ACTIVE_ADC_TILES
                for n in range(self.MAX_ADC_TILES):
                    if (bitvector & 0x1):
                        self.rfdc.adc_tiles[n].SetupFIFOBoth(toggleValue)
                    bitvector = bitvector >> 1

    def sync_tiles(self, dac_target=-1, adc_target=-1, dac=True, adc=True):
        """Synchronize all the active ADC and DAC tiles in the design

        Parameters
        ----------
        dac_target : int
            Set a target latency for the DAC tiles between 0 and 127 cycles passing -1 allows the
            MTS engine to select a latency
        adc_target : int
            Set a target latency for the ADC tiles between 0 and 127 cycles passing -1 allows the
            MTS engine to select a latency
        dac : Boolean
            Enables MTS sync on the DAC
        adc : Boolean
            Enables MTS sync on the ADC
        """
        self._sync_tiles(dac_target, adc_target, dac, adc)
        import mkidgen3.quirks
        if mkidgen3.quirks.MTS.double_sync:
            self._sync_tiles(dac_target, adc_target, dac, adc)

    def _sync_tiles(self, dac_target=-1, adc_target=-1, dac=True, adc=True):
        """ Configures RFSoC MTS alignment"""
        # Set which RF tiles use MTS and turn MTS off
        if self.ACTIVE_ADC_TILES > 0 and adc:
            self.rfdc.mts_adc_config.Tiles = self.ACTIVE_ADC_TILES
            self.rfdc.mts_adc_config.SysRef_Enable = 1
            self.rfdc.mts_adc_config.Target_Latency = adc_target
            self.rfdc.mts_adc()
        elif self.ACTIVE_ADC_TILES == 0:
            self.rfdc.mts_adc_config.Tiles = 0x0
            self.rfdc.mts_adc_config.SysRef_Enable = 0
        if self.ACTIVE_DAC_TILES > 0 and dac:
            self.rfdc.mts_dac_config.Tiles = self.ACTIVE_DAC_TILES  # group defined in binary 0b1111
            self.rfdc.mts_dac_config.SysRef_Enable = 1
            self.rfdc.mts_dac_config.Target_Latency = dac_target
            self.rfdc.mts_dac()
        elif self.ACTIVE_DAC_TILES == 0:
            self.rfdc.mts_dac_config.Tiles = 0x0
            self.rfdc.mts_dac_config.SysRef_Enable = 0

    def set_gain(self, adc_gains=(1.0, 1.0), dac_gains=(1.0,1.0)):
        getLogger(__name__).info('Pretending to set gain, not currently supported')
        return

    def set_qmc(self, adc=None, dac=None, gain=0.0, offset=0, phase=0.0):
        """
        Sets the quadrature error modulation circuit for the specified adc/dac.

        adc: tuple describing (adc tile, adc block) allowed values: 0,1,2,3
        dac: tuple describing (dac tile, dac block) allowed values: 0,1,2,3
        gain: number between 0 and 2 describing data converter gain.
        offset: xxxxx
        phase: xxxxxx

        Example Usage: set_qmc(adc=(0,1), gain=1.5)
        """
        getLogger(__name__).info('Pretending to set QMC, not currently supported')
        return

        settings = {'EnableGain': 1 if gain else 0, 'EnablePhase': 1 if phase else 0, 'EventSource': 0,
                    'GainCorrectionFactor': gain, 'OffsetCorrectionFactor': offset, 'PhaseCorrectionFactor': phase}

        import xrfdc

        if adc is not None:
            self.rfdc.adc_tiles[adc[0]].blocks[adc[1]].QMCSettings = settings
            self.rfdc.adc_tiles[adc[0]].blocks[adc[1]].UpdateEvent(xrfdc.EVENT_QMC)
            print(f"Setting ADC Tile {adc[0]}, Block {adc[1]}")

        if dac is not None:
            self.rfdc.dac_tiles[dac[0]].blocks[dac[1]].QMCSettings = settings
            self.rfdc.dac_tiles[dac[0]].blocks[dac[1]].UpdateEvent(xrfdc.EVENT_QMC)
            print(f"Setting DAC Tile {dac[0]}, Block {dac[1]}")
        return (settings)



from logging import getLogger

try:
    import pynq
    from .drivers import *
    from .equipment_drivers import *
    from .opfb import opfb_bin_number, opfb_bin_center, quantize_frequencies
    from .drivers.ddc import tone_increments
    from . import util
    from . import quirks
    from . import registers
    from .overlay_helpers import *
except ImportError:
    getLogger(__name__).info('pynq not available, functionality will be limited.')
except OSError as e:
    if 'libxrt_core' in str(e):
        getLogger(__name__).info('pynq not available, functionality will be limited.')



import time
from logging import getLogger
from pynq import DefaultIP
import numpy as np


class AxisFIFO(DefaultIP):
    """Support for an AXI FIFO without cut-through support, could be enhanced"""
    bindto = ['xilinx.com:ip:axi_fifo_mm_s:4.2']

    def __init__(self, description):
        super().__init__(description=description)
        self.length = 512

    def reset_tx_fifo(self):
        self.register_map.TDFR = 0x000000A5
        while not self.register_map.ISR.TRC:
            print('Waiting on tx reset complete...')
            time.sleep(1)

    def tx(self, data, destination=0, last_bytes=4, wait=True, check_vacancy=True):
        """
        Data must be an array of uint32

        The AXI FIFO writes the samples as written, so you can't pack pairs of 16 into 32 unless you have a stream data
        width converter and enable TKEEP.
        """
        if check_vacancy and data.size > self.tx_vacancy:
            raise ValueError('Insufficient room in fifo for data')

        getLogger(__name__).debug(f'ISR at TX start: {repr(self.register_map.ISR)}')
        self.register_map.ISR = 0xFFFFFFFF  # Write to clear reset done interrupt bits
        self.register_map.IER.TPOE = 1  # Interrupt if we try to load too much data (should not be possible)
        self.register_map.IER.TSE = 1  # Interrupt on transmit size errors
        self.register_map.IER.TCE = 1  # Enable transmit complete interrupt
        self.register_map.TDR.TDEST = destination  # Transmit Destination address

        for x in data:
            self.mmio.write(self.register_map.TDFD.address, int(x))  # Write value

        self.register_map.TLR.TXL = (data.size - 1) * 4 + last_bytes
        if wait:
            from ..interrupts import ThreadedPLInterruptManager
            _, event = ThreadedPLInterruptManager.get_monitor(self, id=repr(self)+'tx')
            event.wait()
            event.clear()

    def rx(self):
        """Pull all the data out of the FIFO"""
        if not self.register_map.ISR.RC:  # receive is complete
            return None
        self.register_map.ISR = 0xFFFFFFFF  # Write to clear reset done interrupt bits
        getLogger(__name__).debug(f'ISR at RX start: {repr(self.register_map.ISR)}')

        addr = self.register_map.RDFD.address
        occ = self.rx_occupancy
        data = []
        for _ in range(occ):
            data.append(self.mmio.read(addr))
        occ = self.rx_occupancy
        for _ in range(occ):
            data.append(self.mmio.read(addr))
        if self.rx_occupancy:
            getLogger()
        return np.array(data)

    def powerup(self):
        assert self.register_map.ISR == 0x01D00000  # Read interrupt status register (indicates transmit reset complete
        # and receive reset complete)
        self.register_map.ISR = 0xFFFFFFFF  # Write to clear reset done interrupt bits

    @property
    def tx_vacancy(self):
        return self.register_map.TDFV.Vacancy

    @property
    def rx_occupancy(self):
        return self.register_map.RDFO.Occupancy



import logging

import numpy as np
from fpbinary import FpBinary
from pynq import DefaultIP
from pynq.mmio import MMIO
from mkidgen3.fixedpoint import fp_factory
from mkidgen3.opfb import opfb_bin_number, opfb_bin_center, quantize_frequencies
import time
import mkidgen3.opfb as dsp
from logging import getLogger
def tone_increments(freq, quantize=True, **kwargs):
    """
    Compute the DDS tone increment for each frequency (in Hz),
    assumes channel will use OPFB bin returned by mkidgen3.drivers.bintores.opfb_bin_number
    when computing central frequency

    If quantize, the tone increment frequencies will be quantized via dsp.quantize_frequencies
    """
    centers = opfb_bin_center(opfb_bin_number(freq, ssr_raw_order=True), ssr_order=True)
    # This must be 2MHz NOT 2.048MHz, the sign matters! Use 1MHz as that corresponds to ±Pi
    x = (freq - centers)
    if quantize:
        x = quantize_frequencies(x, **kwargs)
    return x / 1e6


# The DDC tone table registers are arranged with least significant bits and bytes at
# lower addresses in 256bit words of p0_7 ... p0_0 i_7 ... i_0
# the increments i are 11 bits and the phase offsets 21 bits.
#
# Addrresses are written by read and write which rea and write 32 bit words.
# so the first 32bit word of the tone table consists of 10 bits of i2, i1 and i0:  i2_9 ... i2_0 i1 i0
# .to_bytes(4, 'big', signed=False) will create a string that prints like this is written, but do not match how
# mimo.write writes the data.
#
# bitstruct.compile('>u11'*8).pack(*[1021]*8) will also printlike this is written, but it too is wrong
# bitstruct.compile('>u11'*8+'<').pack(*[1021]*8)
#
# mimo.write(addr, b'\x??\x??\x??\x??...')  writes to the core as
# B0 B1 B2 .... The key here is that \x## is read as the number 0x00, so 0xF0 is 240 and 0x0F is 15. So we are writing the numbers least significant byte in first (and least significant bit as well) and will read from the core in the same manner.
#
# bitstruct does not seem to be able to properly render this
# bitstruct.pack('>u32<', x) will properly pack a u32 into bytes but fails with u11


class DDC(DefaultIP):
    offset_tones = 0x2000
    TONE_FORMAT = (1, 10, 'signed')  # ap_fixed<11,1>
    PHASE0_FORMAT = (1, 20, 'signed')  # ap_fixed<21,1>

    bindto = ['mazinlab:mkidgen3:resonator_dds:1.33']

    def __init__(self, description):
        """
        The core uses an array of 256 values, each consisting of 8 32 bit numbers packed into 256 bit word that
        specifies the phase offset and phase increment used to digitally down-convert the corresponding resonator
        channel. Each 32 bit number is itself a packed fixed point number with 1 integer bit, 21 bits for the phase and
        11 for the tone. The high bits are for the phase offset.

        0x2000 ~
        0x3fff : Memory 'tones' (256 * 256b)  inc0-8 p0 0-8
                 Word 8n   : bit [31:0] - tones[n][31: 0]
                 Word 8n+1 : bit [31:0] - tones[n][63:32]
                 Word 8n+2 : bit [31:0] - tones[n][95:64]
                 Word 8n+3 : bit [31:0] - tones[n][127:96]
                 Word 8n+4 : bit [31:0] - tones[n][159:128]
                 Word 8n+5 : bit [31:0] - tones[n][191:160]
                 Word 8n+6 : bit [31:0] - tones[n][223:192]
                 Word 8n+7 : bit [31:0] - tones[n][255:224]


        """
        super().__init__(description=description)

    @staticmethod
    def _checkgroup(group_ndx):
        if group_ndx < 0 or group_ndx > 255:
            raise ValueError('group_ndx must be in [0,255]')

    def read_group(self, group_ndx, raw=False):
        """Read the numbers in the group from the core and convert them from binary data to python numbers"""
        self._checkgroup(group_ndx)

        tone_fmt = fp_factory(*self.TONE_FORMAT, frombits=True)
        phase_fmt = fp_factory(*self.PHASE0_FORMAT, frombits=True)

        t_bits = sum(self.TONE_FORMAT[:2])
        p_bits = sum(self.PHASE0_FORMAT[:2])

        t_mask = 2 ** t_bits - 1
        p_mask = 2 ** p_bits - 1

        x = 0
        for i in range(8):
            v = self.read(self.offset_tones + 32 * group_ndx + i * 4)
            x |= v << (32 * i)

        tones = [(x >> (t_bits * i)) & t_mask for i in range(8)]
        x >>= 88
        phases = [(x >> (p_bits * i)) & p_mask for i in range(8)]

        if not raw:
            tones = [float(tone_fmt(v)) for v in tones]
            phases = [float(phase_fmt(v)) for v in phases]

        return tones, phases

    def write_group(self, group_ndx, increments, phases, raw=False):
        """ Convert the numbers in the group from python data to binary data and load it into the core """
        self._checkgroup(group_ndx)
        if len(increments) != 8 or len(phases) != 8:
            raise ValueError('len(group)!=8')

        tone_fmt = (lambda x: x) if raw else fp_factory(*self.TONE_FORMAT, include_index=True)
        phase_fmt = (lambda x: x) if raw else fp_factory(*self.PHASE0_FORMAT, include_index=True)

        t_bits = sum(self.TONE_FORMAT[:2])
        p_bits = sum(self.PHASE0_FORMAT[:2])

        inc = 0
        for i, v in enumerate(map(tone_fmt, increments)):
            inc |= v << (t_bits * i)

        pha = 0
        for i, v in enumerate(map(phase_fmt, phases)):
            pha |= v << (p_bits * i)

        d = (pha << 88) | inc
        data = d.to_bytes(32, 'little', signed=False)
        self.write(self.offset_tones + 32 * group_ndx, data)

    def pack_group(self, increments, phases, raw=False):
        """ Convert the numbers in the group from python data to binary data and load it into the core """

        if len(increments) != 8 or len(phases) != 8:
            raise ValueError('len(group)!=8')

        tone_fmt = (lambda x: x) if raw else fp_factory(*self.TONE_FORMAT, include_index=True)
        phase_fmt = (lambda x: x) if raw else fp_factory(*self.PHASE0_FORMAT, include_index=True)

        t_bits = sum(self.TONE_FORMAT[:2])
        p_bits = sum(self.PHASE0_FORMAT[:2])

        inc = 0
        for i, v in enumerate(map(tone_fmt, increments)):
            inc |= v << (t_bits * i)

        pha = 0
        for i, v in enumerate(map(phase_fmt, phases)):
            pha |= v << (p_bits * i)

        d = (pha << 88) | inc
        data = d.to_bytes(32, 'little', signed=False)
        data = np.frombuffer(data, dtype=np.uint32)
        return data

    @property
    def tones(self):
        return np.hstack([self.read_group(g) for g in range(256)])

    @tones.setter
    def tones(self, tones):
        """tones is a [2,2048] array of tone increments and phase offsets """
        if tones.shape != (2, 2048):
            raise ValueError('tones.shape !=(2,2048)')
        if tones.min() < -1 or tones.max() >= 1:
            raise ValueError('Tones must be in [-1,1)')
        for i in range(256):
            self.write_group(i, *tones[:, i * 8:i * 8 + 8])


class CenteringDDC(DDC):
    CENTER_FORMAT = (1, 15, 'signed')  # ap_fixed<16,15>
    offset_centers = 0x4000
    bindto = ['mazinlab:mkidgen3:resonator_ddc_control:1.0']

    def reset_accumulator(self):
        """ Reset any accumulated phase """
        self.register_map.clear_accumulator = True
        time.sleep(256*4/512e6)
        self.register_map.clear_accumulator = False

    @property
    def centers(self):
        """ Returns an array of 2048 complex loop centers [1,1] """
        mmio = MMIO(self.offset_centers, length=4 * 2048)
        u32d = np.array(mmio.array, dtype=np.uint32)
        u16 = np.frombuffer(u32d, dtype=np.uint16).reshape((2048, 2))
        center_fmt = fp_factory(*self.CENTER_FORMAT, frombits=True, include_index=True)
        data = np.zeros(2048, dtype=np.complex64)
        data.real = [float(center_fmt(int(x))) for x in u16[:, 0]]
        data.imag = [float(center_fmt(int(x))) for x in u16[:, 1]]
        return data

    @centers.setter
    def centers(self, centers):
        """ Centers is an array of 2048 complex loop centers [1,1] """
        if centers.shape != (2048,):
            raise ValueError('centers.shape != (2048,)')
        if np.abs(centers.real).max() > 1 or np.abs(centers.imag).max() > 1:
            raise ValueError('Centers must be in [-1,1)')
        if np.abs(centers).max() > 1:
            logging.getLogger(__name__).warning('Centers contains magnitudes outside of the unit circle')

        center_fmt = fp_factory(*self.CENTER_FORMAT, frombits=False, include_index=True)

        data = np.zeros((2048, 2), dtype=np.uint16)
        data[:, 0] = [center_fmt(x.real) for x in centers]
        data[:, 1] = [center_fmt(x.imag) for x in centers]
        u32d = np.frombuffer(data, dtype=np.uint32)
        mmio = MMIO(self.offset_centers, length=4 * 2048)
        mmio.array[:] = u32d


class ThreepartDDC(CenteringDDC):
    def __init__(self, bram_controller_mmio):
        #backup: init with bram_controller_mmio = MMIO(offset, length=8 * 2048)
        self.mmio=bram_controller_mmio

    def configure(self, tones=None, phase_offset=None, loop_center=None, center_relative=False, quantize=True):
        """
        Configure the DDC to down-convert resonator channels containing the specified tones.

        Optionally phase offsets may be specified for each channel. Phase offsets are in [-pi,pi] radians. Values outside
        are clipped.

        Optionally loop centers may be specified (complex nominally [-1,1) though values will be clipped when converted
        from floating to fix point).

        Center relative means specify the frequency relative to the bin center.

        Only the first 2048 frequencies/offsets will be used. DDC tones are assigned in the order frequencies are
        specified. If fewer than 2048 frequencies are specified no assumptions may be made about the DDC settings for the
        remaining channels, however they may be determined by inspecting the .tones property of resonator_ddc.
        """
        data = np.zeros((2, 2048))
        freq = np.asarray(tones)

        if freq.size > 2048:
            getLogger(__name__).warning(f'Using first 2048 of {freq.size} provided frequencies')

        if center_relative:
            freq = dsp.quantize_frequencies(freq) if quantize else freq
            data[0, :min(freq.size, 2048)] = freq[:2048] / 1e6
        else:
            data[0, :min(freq.size, 2048)] = tone_increments(freq[:2048], quantize=quantize)

        if phase_offset is not None:
            phase_offset = np.asarray(phase_offset)
            if freq.size != phase_offset.size:
                raise ValueError('If provided, phase_offsets must match frequencies')
            phase_offset = (phase_offset / np.pi).clip(-1, 1)
            data[1, :min(freq.size, 2048)] = phase_offset[:2048]

        centers = np.zeros(2048, dtype=np.complex64)
        if loop_center is not None:
            loop_center = np.asarray(loop_center)
            if freq.size != loop_center.size:
                raise ValueError('If provided, loop_center must match frequencies')
            centers[:loop_center[:2048].size] = loop_center[:2048]
            if (np.abs(centers) > 1).any():
                getLogger(__name__).warning(f'Loop centers exist outside of the unit circle')

        self._configure(data, centers)

    def _configure(self, tones, centers):
        """ Centers is an array of 2048 complex loop centers [1,1] """
        if centers.shape != (2048,):
            raise ValueError('centers.shape != (2048,)')
        if np.abs(centers.real).max() > 1 or np.abs(centers.imag).max() > 1:
            raise ValueError('Centers must be in [-1,1)')
        if np.abs(centers).max() > 1:
            logging.getLogger(__name__).warning('Centers contains magnitudes outside of the unit circle')

        if tones.shape != (2, 2048):
            raise ValueError('tones.shape !=(2,2048)')
        if tones.min() < -1 or tones.max() >= 1:
            raise ValueError('Tones must be in [-1,1)')

        center_fmt = fp_factory(*self.CENTER_FORMAT, frombits=False, include_index=True)

        data2 = np.zeros((2048, 2), dtype=np.uint16)
        data2[:, 0] = [center_fmt(x.real) for x in centers]
        data2[:, 1] = [center_fmt(x.imag) for x in centers]

        data1 = np.zeros((256, 8), dtype=np.uint32)
        for i in range(256):
            data1[i, :] = self.pack_group(*(-tones[:, i * 8:i * 8 + 8]))

        data = np.zeros((256, 2, 8), dtype=np.uint32)
        data[:, 0, :] = data1
        data[:, 1, :] = np.frombuffer(data2, dtype=np.uint32).reshape((256, 8))
        u32d = np.frombuffer(data, dtype=np.uint32)
        # mmio = MMIO(self.offsoffsetet, length=8 * 2048)
        self.mmio.array[:u32d.size] = u32d



import subprocess
import numpy as np

MAX_CAP_RAM_BYTES = 2**32
PL_DDR4_ADDR = 0x500000000
N_IQ_GROUPS = 256

PHOTON_DTYPE = np.dtype([('time', np.uint64), ('phase', np.int16), ('id', np.uint16)])

def unpack_photons(x, out=None, n=0):
    """
    Unpack packed photons, optionally accumulating them into an existing output array

    Args:
        x: an array of packed photons
        out: optional, an array of type PhotonMAXI.PHOTON_DTYPE with the shape of x, but see n
        n: optional, an index into out to insert unpacked photons, the first axis is used if >1d an
            IndexError is raised if there is insufficient space.

    Returns: the unpacked photon array

    """
    if out is None:
        n = 0
    elif x.shape[0] + n > out.shape[0]:
        raise IndexError('Output array is too small')

    ret = np.zeros(x.shape, dtype=PHOTON_DTYPE) if out is None else out
    sl = slice(n, n + x.shape[0])
    ret['phase'][sl] = x & 0xffff
    ret['time'][sl] = x >> 28
    ret['id'][sl] = (x >> 16) & 0xfff
    return ret


def get_board_name():
    x = subprocess.run(['cat', '/proc/device-tree/chosen/pynq_board'], capture_output=True, text=True).stdout
    return x.strip().strip('\x00')


def enable_axi_timeout():
    """ See https://discuss.pynq.io/t/help-debuging-chronic-pynq-system-hang/970"""
    import pynq
    #LPD
    mmio = pynq.MMIO(0xFF416000, 64)
    mmio.write(0x18, 3)  # Return slave errors when timeouts occur
    mmio.write(0x20, 0x1020)  # Set and enable prescale of 32 which should be about 10 ms
    mmio.write(0x10, 0x3)  # Enable transactions tracking
    mmio.write(0x14, 0x3)  # Enable timeouts

    #FPD
    mmio = pynq.MMIO(0xFD610000, 64)
    mmio.write(0x18, 7)  # Return slave errors when timeouts occur
    mmio.write(0x20, 0x1020)  # Set and enable prescale of 32 which should be about 10 ms
    mmio.write(0x10, 0x7)  # Enable transactions tracking
    mmio.write(0x14, 0x7)  # Enable timeouts


def get_pldram_addr(hwhpath):
    """Return PL DRAM start address as specified in hwh"""
    pldram_addr = None
    pldramstr = '<MEMRANGE ADDRESSBLOCK="C0_DDR4_ADDRESS_BLOCK" BASENAME="C_BASEADDR" BASEVALUE="'
    with open(hwhpath, "r") as hwh:
        for line in hwh:
            if pldramstr in line:
                break
        try:
            pldram_addr = hex(int(line[88:99], 16))
        except LookupError:
            print('PL DRAM not found')
    return pldram_addr


def dma_status(dma):
    # dma.recvchannel.idle,dma.sendchannel.idle
    msg = ("DMA:\n"
           f" Buffer Length: {dma.buffer_max_size} bytes\n"
           " MM2s\n"
           f" Idle:{dma.sendchannel.idle}\n"
           f" MM2S_DMASR (status):{hex(dma.mmio.read(4))}\n"
           f" MM2S_SA (ptr) :{hex(dma.mmio.read(24))}\n"
           f" MM2S_LENGTH (len):{dma.mmio.read(40)}\n"
           " S2MM\n"
           f" Idle:{dma.recvchannel.idle}\n"
           f" S2MM_DMASR (status):{hex(dma.mmio.read(52))}\n"
           f" S2MM_DA (ptr) :{hex(dma.mmio.read(72))}\n"
           f" S2MM_LENGTH (len):{dma.mmio.read(88)}")
    print(msg)


# LUT of property addresses for our data-driven properties
_qpsk_props = [("transfer_symbol", 0), ("transfer_fft", 4),
               ("transfer_time", 60), ("reset_symbol", 8), ("reset_fft", 12),
               ("reset_time", 48), ("packetsize_symbol", 16),
               ("packetsize_rf", 20), ("packetsize_fft", 24),
               ("packetsize_time", 52), ("autorestart_symbol", 36),
               ("autorestart_fft", 40), ("autorestart_time", 56),
               ("lfsr_rst", 28), ("enable", 32), ("output_gain", 44)]


# Func to return a MMIO getter and setter based on a relative addr


def _mimo_attacher(class_def, mimo_regs):
    # Generate getters and setters based on mimo_regs
    def _create_mmio_property(addr):
        def _get(self):
            return self.read(addr)

        def _set(self, value):
            self.write(addr, value)

        return property(_get, _set)

    for (name, addr) in mimo_regs:
        setattr(class_def, name, _create_mmio_property(addr))


def check_description_for(description, kinds, check_version=False, force_dict=False):
    if isinstance(kinds, str):
        kinds = (kinds,)
    ret = {k: [] for k in kinds}
    for k in description['ip']:
        kind = description['ip'][k].get('type', '')
        if not check_version:
            kind, _, version = kind.rpartition(':')
        if kind in kinds:
            ret[kind].append(k)

    return ret if force_dict or len(kinds)>1 else ret[kinds[0]]


def print_plstatus():
    from pynq import PL
    print(f"PL Bitfile: {PL.bitfile_name}\nPL Timestamp: {PL.timestamp}\n")


class DummyOverlay:
    class DummyBuffer(np.ndarray):
        def freebuffer(self):
            pass

    class DummyCap:
        @staticmethod
        def capture(csize, *args, **kwargs):
            n = csize * 2
            return np.random.uniform(low=-10000, high=10000, size=n).astype(np.int16).view(DummyOverlay.DummyBuffer)

        @staticmethod
        def ready():
            return True

    def __init__(self, bitstream, *args, **kwargs):
        self.capture = DummyOverlay.DummyCap()



import numpy as np
import pynq.buffer as buffer
from pynq import allocate, DefaultIP, DefaultHierarchy
import time
import asyncio
from logging import getLogger
from typing import Iterable

from ..mkidpynq import N_IQ_GROUPS, MAX_CAP_RAM_BYTES, PL_DDR4_ADDR, \
    check_description_for  # config, overlay details
from ..system_parameters import N_CHANNELS, N_IQ_GROUPS, N_PHASE_GROUPS, channel_to_iqgroup, channel_to_phasegroup, iqgroup_to_channel, phasegroup_to_channel, ADC_INPUT_WARN
from ..util import ps_ram_sane, format_bytes
from ..interrupts import ThreadedPLInterruptManager


class FilterIQ(DefaultIP):
    """
    """
    bindto = ['mazinlab:mkidgen3:filter_iq:0.1']
    ADDR_KEEP = 0x10  # 8 words
    ADDR_LASTGRP = 0x34  # 8 bits

    def __init__(self, description):
        super().__init__(description=description)

    @property
    def n_kept(self):
        """ Return the number of groups being preserved"""
        return len(self.keep)

    @property
    def keep(self):
        """
        Get the IQ groups that are being preserved. Returns an iterable of resonator channel group numbers (0-255).

        e.g. i in keep => resonator channels i*8 ... i*8+7 are preserved
        """
        ret = []
        for i in range(8):
            k = self.read(self.ADDR_KEEP + 0x4 * i)
            ret += [j + 32 * i for j in range(32) if k & (1 << j)]
        return ret

    def keep_channels(self, channels):
        """Like keep= but assume channels"""
        self.keep = channel_to_iqgroup(channels)

    @keep.setter
    def keep(self, groups: Iterable):
        """
        Tell the block to preserve IQ groups. Set to an iterable of resonator channel IQ group numbers (0-255). If
        iterable contains values > 255 it is assumed numbers indicate resonator channels and the necessary groups
        will be computed. 'all' may be used as a shortcut.

        IQs are processed in 256 groups of 8, so to capture the IQ values of resonator channels 0, 37,
        and 2047 groups must be set to either 'all' or should
        include e.g. (0, 4, 255). Note that this will cause IQs for channels 0-7, 32-39, and 2040-2047 to be captured.

        len(groups) must be a multiple of 2
        """

        # Determine keep, keep a group if all or if the group number is in groups
        keep = np.zeros(8, dtype=np.uint32)
        if isinstance(groups, str):
            if groups.lower() != 'all':
                raise ValueError("The only legal string for keep is 'all'")
            keep += 0xFFFFFFFF
            last = N_IQ_GROUPS-1
        else:
            groups = set(groups)

            if len(groups) % 2:
                groups.add(set(range(256)).difference(groups).pop())
            last = int(max(groups))
            if max(groups) > N_IQ_GROUPS-1 or min(groups) < 0:
                raise ValueError(f'Groups must be in range 0-{N_IQ_GROUPS-1}')

            for g in groups:
                i = g // 32  # (0-7)
                keep[i] |= (1 << (g % 32))  # set the correct bit

        for i, k in enumerate(keep):
            self.write(self.ADDR_KEEP + 0x4 * i, int(k))
        self.write(self.ADDR_LASTGRP, last)


class FilterPhase(DefaultIP):
    bindto = ['mazinlab:mkidgen3:filter_phase:0.2']
    ADDR_KEEP = 0x10  # 4 words
    ADDR_LASTGRP = 0x24  # 7 bits

    def __init__(self, description):
        super().__init__(description=description)

    @property
    def n_kept(self):
        """ Return the number of groups being preserved"""
        return len(self.keep)

    @property
    def keep(self):
        """
        Get the phase groups that are being preserved. Returns an iterable of resonator channel group numbers (0-127).

        e.g. i in keep => resonator channels i*16 ... i*16+15 are preserved
        """
        ret = []
        for i in range(4):
            k = self.read(self.ADDR_KEEP + 0x4 * i)
            ret += [j + 32 * i for j in range(32) if k & (1 << j)]
        return ret

    def keep_channels(self, channels):
        """Like keep= but assume channels"""
        self.keep = channel_to_phasegroup(channels)

    @keep.setter
    def keep(self, groups):
        """
        Tell the block to preserve phase groups. Set to an iterable of resonator channel phase group numbers (0-129). If
        iterable contains values > 127 it is assumed numbers indicate resonator channels and the necessary groups
        will be computed. 'all' may be used as a shortcut.

        Phases are captured in 128 groups of 16, so to capture the values of resonator channels 0, 37,
        and 2047 groups must be set to either 'all' or should
        include (0, 2, 127). Note that this will cause phases for channels 0-16, 32-48, and 2032-2047 to be captured.
        """

        # Determine keep, keep a group if all or if the group number is in groups
        keep = np.zeros(4, dtype=np.uint32)
        if isinstance(groups, str):
            if groups.lower() != 'all':
                raise ValueError("The only legal string for keep is 'all'")
            keep += 0xFFFFFFFF
            last = N_PHASE_GROUPS-1
        else:
            if len(groups) % 2:
                groups.add(set(range(512)).difference(groups).pop())
            last = int(max(groups))

            if max(groups) > N_PHASE_GROUPS-1 or min(groups) < 0:
                raise ValueError(f'Groups must be in range 0-{N_PHASE_GROUPS-1}')

            for g in groups:
                i = g // 32  # (0-7)
                keep[i] |= (1 << (g % 32))  # set the correct bit

        for i, k in enumerate(keep):
            self.write(self.ADDR_KEEP + 0x4 * i, int(k))
        self.write(self.ADDR_LASTGRP, last)


class CaptureHierarchy(DefaultHierarchy):
    """
    Capture Hierarchy.

    Driver supports capture at up to five locations. They are numbered 1-5 in the source map in accordance with their
    position on the axis switch. TODO: Need to standardize these keys and names
    switch port: tap_name: description:
    0: 'adc': connected to the output of the adc
    1: 'rawiq': connected to the output of bin to res, can be used to look at the OPFB bins before the DDC
    2: 'ddciq': connected to the output of the DDC before the lowpass
    3: 'filtphase': connected to the output of matched filters
    4: 'debugiq': placeholder for a filter_iq IP connected to 4th switch port
    5. 'debugpphase' placeholder for a filter_phase IP connected to 5th switch port
    """
    IQ_MAP = {'rawiq': 0, 'ddciq': 1, 'debugiq': 2}  #tap name MUST include 'iq' and shall not include 'phase'
    PHASE_MAP = {'filtphase': 0, 'debugphase': 1} #tap name MUST include 'phase' and shall not include 'iq'
    SOURCE_MAP = dict(adc=0, rawiq=1, ddciq=2, filtphase=3, debugiq=4, debugphase=5)
    USE_CACHEABLE_BUFFERS = True

    def __init__(self, description):
        super().__init__(description)
        self.filter_iq = {}
        self.filter_phase = {}

        switchloc = getattr(self, 'Switchboard', None) or getattr(self, 'switchboard', None) or self
        self.switch = getattr(switchloc, 'axis_switch_0', None) or getattr(switchloc, 'axis_switch', None)

        if not hasattr(self, 'axis2mm'):
            self.axis2mm = self.axis2mm_0

        self._fetch_filter_blocks()

    @staticmethod
    def checkhierarchy(description):
        found = check_description_for(description, ('xilinx.com:axis_switch', 'mazinlab:mkidgen3:filter_iq',
                                                    'xilinx.com:module_ref:axis2mm'))
        return description['fullpath'] == 'capture' and bool(len(found['xilinx.com:module_ref:axis2mm']))

    def _fetch_filter_blocks(self):
        i = 0
        while True:
            try:
                self.filter_iq[i] = getattr(self, f'filter_iq_{i}')
                i += 1
            except AttributeError:
                break
        if not i:
            getLogger(__name__).warning(f'Capture does not support iq capture')
        else:
            getLogger(__name__).debug(f'Found {i} iq capture taps')

        for n, i in self.IQ_MAP.items():
            self.filter_iq[n] = self.filter_iq.get(i, None)

        i = 0
        while True:
            try:
                self.filter_phase[i] = getattr(self, f'filter_phase_{i}')
                i += 1
            except AttributeError:
                break
        if not i:
            getLogger(__name__).warning(f'Capture does not support phase capture')
        else:
            getLogger(__name__).debug(f'Found {i} phase capture taps')

        for n, i in self.PHASE_MAP.items():
            self.filter_phase[n] = self.filter_phase.get(i, None)

        base = pbase = 1
        iqs = list(self.IQ_MAP.items())
        iqs.sort(key=lambda x: x[1])
        for k, v in iqs:
            if self.filter_iq.get(k, None):
                self.SOURCE_MAP[k] = v + base
                pbase = v + base

        ph = list(self.PHASE_MAP.items())
        ph.sort(key=lambda x: x[1])
        for k, v in ph:
            self.SOURCE_MAP[k] = v + pbase + 1

    def flush(self, n):
        """Capture n*64 bytes to flush a fifo"""
        buffer = allocate(n, dtype='u64', target=self.ddr4_0)
        self.axis2mm.addr = buffer.device_address
        self.axis2mm.len = 64 * n
        self.axis2mm.start(continuous=False, increment=True)
        del buffer

    def looping_capture(self, source: str, n: int, buffers: buffer.PynqBuffer, callback):
        getLogger(__name__).warning('Looping capture untested, exercise case')
        if self.switch is not None:
            self.switch.set_driver(slave=self.SOURCE_MAP[source], commit=True)
        if n % 64:
            raise ValueError('Can only capture in multiples of 64 bytes')
        if not self.axis2mm.ready:
            raise IOError("capture core not ready, this shouldn't happen."
                          " Try calling .axis2mm.abort() followed by .axis2mm.clear_error()"
                          " then try a small throwaway capture (data order may not be aligned in the first capture "
                          "after a reset).")

        i = 0
        self.axis2mm.addr = buffers[0].device_address
        self.axis2mm.len = n
        self.axis2mm.start(continuous=True, increment=True)

        while not self.terminate_looped_capture:
            self.wait()
            stat = self.cmd_ctrl_reg
            if stat['r_err']:
                getLogger(__name__).error(f'Aborting capture loop due to error: {stat}')
                break
            else:
                callback(buffers[i])
                i = i + 1 if i < len(buffers) else 0
                self.axis2mm.addr = buffers[i].device_address
                self.axis2mm.len = n
        self.abort()

    def _capture(self, source, n, buffer_addr):
        if self.switch is not None:
            self.switch.set_driver(slave=self.SOURCE_MAP[source], commit=True)
        if n % 64:
            raise ValueError('Can only capture in multiples of 64 bytes')
        if not self.axis2mm.ready:
            raise IOError("capture core not ready, this shouldn't happen."
                          " Try calling .axis2mm.abort() followed by .axis2mm.clear_error()"
                          " then try a small throwaway capture (data order may not be aligned in the first capture "
                          "after a reset).")

        self.axis2mm.addr = buffer_addr
        self.axis2mm.len = n
        _, e = ThreadedPLInterruptManager.get_monitor(self.axis2mm._interrupts['o_int']['fullpath'], id='capheir')
        e.clear()
        getLogger(__name__).debug(f'Starting capture of {format_bytes(n)} ({n // 64} beats) to address {hex(buffer_addr)} from '
                                  f'source {source}: \n Interrupt Status:' +
                                  str(ThreadedPLInterruptManager.get_status(self.axis2mm._interrupts['o_int']['fullpath'])))
        self.axis2mm.start(continuous=False, increment=True)

    def is_ready(self):
        return self.axis2mm.ready

    def keep_channels(self, tap, channels):
        if isinstance(channels, str):
            if channels.lower() == 'all':
                channels = list(range(N_CHANNELS))
            else:
                raise ValueError(f'Invalid channel type: {channels}')
        if tap in self.IQ_MAP:
            return self.filter_iq[tap].keep_channels(channels)
        elif tap in self.PHASE_MAP:
            return self.filter_phase[tap].keep_channels(channels)

    def kept_channels(self, tap):
        if tap in self.IQ_MAP:
            return iqgroup_to_channel(self.filter_iq[tap].keep)
        elif tap in self.PHASE_MAP:
            return phasegroup_to_channel(self.filter_phase[tap].keep)
        else:
            return (1,)

    def capture(self, n, tap, groups='all', wait=True):
        """
        Do a capture to PL Ram
        Args:
            n: the number of samples to capture, will be truncated to the nearest capturable amount.
            tap: an iq, phase, or adc capture tap location. 'adc' or see IQ_MAP and PHASE_MAP keys. Raises value error
            if invalid.
            groups: Which sample groups to capture (only relevant for iq and phase)
            wait: wait for the capture to complete. If not waiting it is necessary to call axis2mm.errors()
            to check for any capture errors. If there are errors axis2mm.clear_errors() will be called.

        Returns: The capture buffer or a dictionary of capture errors

        """
        try:
            assert (int(n)-n) == 0
            n = int(n)
        except:
            raise TypeError('n must be effectively an integer')

        if tap in self.IQ_MAP:
            buf = self.capture_iq(n, tap_location=tap, duration=False, groups=groups, wait=wait)
        elif tap in self.PHASE_MAP:
            buf = self.capture_phase(n, tap_location=tap, duration=False, groups=groups, wait=wait)
        elif tap == 'adc':
            buf = self.capture_adc(n, complex=False, duration=False, wait=wait)
        else:
            valid = ('adc',) + tuple(self.IQ_MAP.keys()) + tuple(self.PHASE_MAP.keys())
            raise ValueError(f'{tap} is not a valid capture location from {valid}')

        if wait:
            errors = self.axis2mm.errors()
            if errors:
                del buf
                self.axis2mm.clear_error()
                return errors

        return buf

    def capture_iq(self, n, groups='all', tap_location='ddciq', duration=False, wait=True):
        """
        potentially valid tap locations are the keys of CaptureHierarchy.IQ_MAP
        if buffer is None one will be allocated, if groups is None it will not be set
        """
        if duration:
            # n samples = t[ms] * 2e6[samples/sec]
            n = int(np.floor(n * 2e-3 * 1e6))

        if n <= 0:
            raise ValueError('Must request at least 1 sample')

        if self.filter_iq.get(tap_location, None) is None:
            raise ValueError(f'Unsupported IQ capture location: {tap_location}')

        if groups is not None:
            self.filter_iq[tap_location].keep = groups
        n_groups = len(self.filter_iq[tap_location].keep)

        # each group is 8 IQ (32 bytes)
        capture_bytes = n * n_groups * 32

        try:
            buffer = allocate((n, n_groups * 8, 2), dtype='i2', target=self.ddr4_0,
                              cacheable=self.USE_CACHEABLE_BUFFERS if wait else False)
        except RuntimeError:
            getLogger(__name__).warning(f'Insufficient space for requested samples.')
            raise RuntimeError('Insufficient free space')
        addr = buffer.device_address

        # NB this ignores the final bit from a non-128 multiple
        datavolume_mb = capture_bytes / 1024 ** 2
        datarate_mbps = 32 * 512 * n_groups / 256
        captime = datavolume_mb / datarate_mbps

        msg = (f"Capturing {format_bytes(capture_bytes)} of data @ {datarate_mbps:.1f} MiBps. "
               f"ETA {datavolume_mb / datarate_mbps * 1000:.0f} ms")
        getLogger(__name__).debug(msg)

        self._capture(tap_location, capture_bytes, addr)

        if wait:
            if isinstance(wait, bool):
                wait = {}
            if 'duration' not in wait:
                wait['duration'] = captime
            self.wait(**wait)

        buffer.invalidate()

        return buffer

    def capture_adc(self, n, duration=False, complex=False, wait=True):
        """
        samples are captured in multiples of 8 will be clipped as necessary

        Set complex to return a numpy complex array of the result. This implies a copy out of the buffer so care is
        required with memory sizes (i.e. will use 5x standard memory in PS DDR).

        """
        if complex and not wait:
            raise RuntimeError('Complex return not supported with immediate return.')

        if n <= 0:
            raise ValueError('Must request at least 1 sample')

        if duration:
            # n samples = t[ms] * 4.096e9[samples/sec]
            n = int(np.floor(n * 1e-3 * 4.096e9))

        n -= n % 8

        if n <= 0:
            n = 8

        # Compute capturesize, this is the number of adc samples to be written
        capture_bytes = n * 4

        if complex and not ps_ram_sane(n*8):
            raise RuntimeError('Not enough RAM to copy capture to complex')

        try:
            buffer = allocate((n, 2), dtype='i2', target=self.ddr4_0,
                              cacheable=self.USE_CACHEABLE_BUFFERS if wait else False)
        except RuntimeError:
            getLogger(__name__).warning(f'Insufficient space for requested samples.')
            raise RuntimeError('Insufficient free space')
        addr = buffer.device_address

        datavolume_mb = capture_bytes / 1024 ** 2
        datarate_mbps = 32 * 512
        captime = datavolume_mb / datarate_mbps

        msg = (f"Capturing {datavolume_mb:.1f} MB of data @ {datarate_mbps} MB/s. "
               f"ETA {datavolume_mb / datarate_mbps * 1000:.0f} ms")
        getLogger(__name__).debug(msg)

        self._capture('adc', capture_bytes, addr)

        if wait or complex:
            if isinstance(wait, bool):
                wait = {}
            if 'duration' not in wait or (complex and wait['duration'] < captime):
                wait['duration'] = captime
            self.wait(**wait)

        buffer.invalidate()

        if complex:
            d = np.zeros(n, dtype=np.complex64)
            d.real[:] = buffer[:, 0]
            d.imag[:] = buffer[:, 1]
            del buffer
            return d
        else:
            return buffer

    def wait(self, duration=0, interrupt=False):
        """Wait for a capture to complete, use interrupt on axis2mm by default, otherwise sleep for duration"""
        if not interrupt:
            while not self.axis2mm.complete:
                time.sleep(duration / 5000)
        else:
            _, e = ThreadedPLInterruptManager.get_monitor(self.axis2mm._interrupts['o_int']['fullpath'], id='capheir')
            e.wait()
            # self.axis2mm.abort()
            # e.clear()

    def capture_phase(self, n, groups='all', duration=False, tap_location='filtphase', wait:(bool,dict)=True):
        """
        samples are captured in multiples of 16 will be clipped ad necessary
        groups is 0-127 or all, None will leave the filter unchanged

        wait is args to selt.wait
        """
        if self.filter_phase.get(tap_location, None) is None:
            getLogger(__name__).error(f'Bitstream does not support phase capture at tap {tap_location}')
            return None

        if duration:
            # n samples = t[ms] * 1e6[samples/sec]
            n = int(np.floor(n * 1e-3 * 1e6))

        if n <= 0:
            raise ValueError('Must request at least 1 sample')

        if groups is not None:
            self.filter_phase[tap_location].keep = groups
        n_groups = len(self.filter_phase[tap_location].keep)

        # each group is 16 phases (32 bytes)
        capture_bytes = n * 2 * n_groups * 16

        try:
            buffer = allocate((n, n_groups * 16), dtype='i2', target=self.ddr4_0,
                              cacheable=self.USE_CACHEABLE_BUFFERS if wait else False)
        except RuntimeError:
            getLogger(__name__).warning(f'Insufficient space for requested samples.')
            raise RuntimeError('Insufficient free space')
        addr = buffer.device_address

        # NB this ignores the final bit from a non-128 multiple
        datavolume_mb = capture_bytes / 1024 ** 2
        datarate_mbps = 32 * 512 / 4 * n_groups / 128  # phases arrive 4@512 so the filter outputs every 4 clocks
        captime = datavolume_mb / datarate_mbps

        msg = (f"Capturing ~{datavolume_mb:.2f} MB of data @ {datarate_mbps:.1f} MBps. "
               f"ETA {datavolume_mb / datarate_mbps * 1000:.0f} ms")
        getLogger(__name__).debug(msg)

        self._capture(tap_location, capture_bytes, addr)

        if wait:
            if isinstance(wait, bool):
                wait = {}
            if 'duration' not in wait:
                wait['duration'] = captime
            self.wait(**wait)

        buffer.invalidate()

        return buffer

    def ddc_compare_cap(self, n_points=1024):
        """A helper function to capture data just after bin2res and after the ddc"""
        x = self.capture_iq(n_points, 'all', tap_location='rawiq')
        riq = np.array(x)
        del x
        x = self.capture_iq(n_points, 'all', tap_location='ddciq')
        iq = np.array(x)
        del x
        riq = riq[..., 0] + riq[..., 1] * 1j
        iq = iq[..., 0] + iq[..., 1] * 1j
        return riq, iq

    def cap_cordic_compare(self, n_points=1024):
        """A helper function to capture data just after the ddc and after matched filters"""
        x = self.capture_iq(n_points, 'all', tap_location='ddciq')
        riq = np.array(x)
        del x
        x = self.capture_phase(n_points, 'all', tap_location='filtphase')
        phase = np.array(x)
        del x
        riq = riq[..., 0] + riq[..., 1] * 1j
        return riq, phase


class _AXIS2MM:
    @property
    def cmd_ctrl_reg(self):
        reg = self.read(0)
        names = ('r_busy', 'r_err', 'r_complete', 'r_continuous', 'r_increment_n',
                 'r_tlast_syncd_n', 'decode_error', 'slave_error', 'overflow_error',
                 'aborting', 'fifo_len', 'abort')
        bits = (31, 30, 29, 28, 27, 26, 25, 24, 23, 22, (20, 16), (15, 8))

        def cvrt(x, bits):
            if isinstance(bits, int):
                return bool(1 & (reg >> bits))
            else:
                return (reg >> bits[1]) & (2 ** (bits[0] - bits[1] + 1) - 1)

        return {k: cvrt(reg, v) for k, v in zip(names, bits)}

    @property
    def addr(self):
        """5 bytes"""
        return self.read(0x10) | ((self.read(0x14) & 0xffff) << 32)

    @addr.setter
    def addr(self, x):
        #         x&=(2**35-1)
        #         self.write(0x10, x&0xffffffff)
        #         self.write(0x14, (x>>32)&0xffffffff)
        self.write(0x10, x.to_bytes(8, 'little'))

    @property
    def len(self):
        return self.read(0x18)

    @len.setter
    def len(self, x):
        """A number of bytes, not beats!"""
        self.write(0x18, x)

    @property
    def ready(self):
        stat = self.cmd_ctrl_reg
        return not stat['r_busy'] and not stat['r_err'] and not stat['aborting']

    @property
    def complete(self):
        return True if ((self.read(0) >> 29) & 1) else False

    def abort(self):
        #         self.write(0, (self.read(0)^0xff00)|0x2600)
        self.write(0, 0x26000000)

    def clear_error(self):
        self.write(0, 0x40000000)

    def errors(self):
        """Return a dictionary of errors if there are errors else None"""
        x = self.cmd_ctrl_reg
        if x['r_err']:
            return {k:x[k] for k in ('r_err', 'decode_error', 'slave_error', 'overflow_error')}
        else:
            return None

    def start(self, continuous=False, increment=True):
        if not self.ready:
            raise IOError('Not ready, need to abort and/or clear errors')
        x = 0x80000000  # start and do not clear error
        x |= continuous << 28
        x |= (not increment) << 27
        self.write(0, x)


class AXIS2MMIP(DefaultIP, _AXIS2MM):
    bindto = ['xilinx.com:module_ref:axis2mm: 1.0']

    def __init__(self, description):
        super().__init__(description=description)


class AXIS2MMHier(DefaultHierarchy, _AXIS2MM):
    def __init__(self, description):
        super().__init__(description)
        self._core = self.S_AXIL
        self.read = self._core.read
        self.mmio = self._core.mmio
        self.write = self._core.write

    @staticmethod
    def checkhierarchy(description):
        try:
            t = description['ip']['S_AXIL']['type']
        except KeyError:
            return False
        return t == 'xilinx.com:module_ref:axis2mm:1.0'



from logging import getLogger
import numpy as np
import os

ISGOOD = 0b1
ISREVIEWED = 0b10
ISBAD = 0

MAX_ML_SCORE = 1
MAX_ATTEN = 100

LOCUT = 1e9

A_RANGE_CUTOFF = 6e9


def parse_lo(lofreq, frequencies=None, sample_rate=2.048e9):
    """  Sets the attribute LOFreq (in Hz) """
    lo = round(lofreq / (2.0 ** -16) / 1e6) * (2.0 ** -16) * 1e6

    try:
        delta = np.abs(frequencies - lo)
    except AttributeError:
        getLogger(__name__).warning('No frequency list yet loaded. Unable to check if LO is reasonable.')
        return lo

    tofar = delta > sample_rate / 2
    if tofar.all():
        getLogger(__name__).warning('All frequencies more than half a sample rate from '
                                    'the LO. LO: {} Delta min: {} Halfsamp: {} )'.format(lo, delta.min(),
                                                                                         sample_rate / 2))
        raise ValueError('LO out of bounds')
    elif tofar.any():
        getLogger(__name__).warning('Frequencies more than half a sample rate from the LO exist')
    return lo


class SweepFile(object):
    def __init__(self, file):
        self.file = file
        self.feedline = None
        self.resIDs = None
        self.wsfreq = None
        self.flag = None
        self.wsatten = None
        self.mlatten = None
        self.mlfreq = None
        self.ml_isgood_score = None
        self.ml_isbad_score = None
        self.phases = None
        self.iqRatios = None
        self.freq = None
        self.atten = None
        self._load()
        self._vet()

    @property
    def goodmlfreq(self):
        return self.mlfreq[self.flag & ISGOOD]

    def sort(self):
        s = np.argsort(self.resIDs)
        self.resIDs = self.resIDs[s]
        self.wsfreq = self.wsfreq[s]
        self.flag = self.flag[s]
        self.mlfreq = self.mlfreq[s]
        self.mlatten = self.mlatten[s]
        self.atten = self.atten[s]
        self.ml_isgood_score = self.ml_isgood_score[s]
        self.ml_isbad_score = self.ml_isbad_score[s]
        self.freq = self.freq[s]

    def toarray(self):
        return np.array([self.resIDs, self.flag, self.wsfreq, self.mlfreq, self.mlatten, self.freq,
                         self.atten, self.ml_isgood_score, self.ml_isbad_score, self.phases, self.iqRatios])

    def lomask(self, lo):
        return ((self.flag & ISGOOD) & (~np.isnan(self.freq)) & (np.abs(self.freq - lo) < LOCUT) & (
                    self.atten > 0)).astype(bool)

    def vet(self):
        if (np.abs(self.atten[~np.isnan(self.atten)]) > MAX_ATTEN).any():
            getLogger(__name__).warning('odd attens')
        if (np.abs(self.ml_isgood_score[~np.isnan(self.ml_isgood_score)]) > MAX_ML_SCORE).any():
            getLogger(__name__).warning('bad ml good score')
        if (np.abs(self.ml_isbad_score[~np.isnan(self.ml_isbad_score)]) > MAX_ML_SCORE).any():
            getLogger(__name__).warning('bad ml bad scores')

        assert self.resIDs.size == np.unique(self.resIDs).size, "Resonator IDs must be unique."

        assert (self.resIDs.size == self.wsfreq.size == self.flag.size ==
                self.atten.size == self.mlfreq.size == self.ml_isgood_score.size ==
                self.ml_isbad_score.size)

    def genheader(self, useSBSup=False):
        if useSBSup:
            header = ('feedline={}\n'
                      'wsatten={}\n'
                      'rID\trFlag\twsFreq\tmlFreq\tmlatten\tfreq\tatten\tmlGood\tmlBad\tphases\tiqratios')
        else:
            header = ('feedline={}\n'
                      'wsatten={}\n'
                      'rID\trFlag\twsFreq\tmlFreq\tmlatten\tfreq\tatten\tmlGood\tmlBad')
        return header.format(self.feedline, self.wsatten)

    def save(self, file='', saveSBSupData=False):
        sf = file.format(feedline=self.feedline) if file else self.file.format(feedline=self.feedline)
        self.vet()
        if saveSBSupData:
            np.savetxt(sf, self.toarray().T, fmt="%8d %1u %16.7f %16.7f %5.1f %16.7f %5.1f %6.4f %6.4f %6.4f %6.4f",
                       header=self.genheader(True))
        else:
            np.savetxt(sf, self.toarray().T[:, :-2], fmt="%8d %1u %16.7f %16.7f %5.1f %16.7f %5.1f %6.4f %6.4f",
                       header=self.genheader(False))

    def _vet(self):

        assert (self.resIDs.size == self.wsfreq.size == self.flag.size == self.atten.size == self.freq.size ==
                self.mlatten.size == self.mlfreq.size == self.ml_isgood_score.size == self.ml_isbad_score.size)

        for x in (self.freq, self.mlfreq, self.wsfreq):
            use = ~np.isnan(x)
            if x[use].size != np.unique(x[use]).size:
                getLogger(__name__).warning("Found non-unique frequencies")

        self.flag = self.flag.astype(int)
        self.resIDs = self.resIDs.astype(int)
        self.feedline = int(self.resIDs[0] / 10000)

    def _load(self):
        d = np.loadtxt(self.file.format(feedline=self.feedline), unpack=True)
        if d.ndim == 1:  # allows files with single res
            d = np.expand_dims(d, axis=1)
        try:
            if d.shape[0] == 11:
                self.resIDs, self.flag, self.wsfreq, self.mlfreq, self.mlatten, \
                    self.freq, self.atten, self.ml_isgood_score, self.ml_isbad_score, self.phases, self.iqRatios = d
            if d.shape[0] == 9:
                self.resIDs, self.flag, self.wsfreq, self.mlfreq, self.mlatten, \
                    self.freq, self.atten, self.ml_isgood_score, self.ml_isbad_score = d
                self.phases = np.full_like(self.resIDs, 0, dtype=float)
                self.iqRatios = np.full_like(self.resIDs, 1, dtype=float)
            elif d.shape[0] == 7:
                self.resIDs, self.flag, self.wsfreq, self.mlfreq, self.mlatten, \
                    self.ml_isgood_score, self.ml_isbad_score = d
                self.freq = self.mlfreq.copy()
                self.atten = self.mlatten.copy()
                self.phases = np.full_like(self.resIDs, 0, dtype=float)
                self.iqRatios = np.full_like(self.resIDs, 1, dtype=float)
            elif d.shape[0] == 5:
                self.resIDs, self.freq, self.atten, self.phases, self.iqRatios = d
                self.wsfreq = self.freq.copy()
                self.mlfreq = self.freq.copy()
                self.mlatten = self.atten.copy()
                self.flag = np.full_like(self.resIDs, ISGOOD, dtype=int)
                self.ml_isgood_score = np.full_like(self.resIDs, np.nan, dtype=float)
                self.ml_isbad_score = np.full_like(self.resIDs, np.nan, dtype=float)
            else:
                self.resIDs, self.freq, self.atten = d
                self.wsfreq = self.freq.copy()
                self.mlfreq = self.freq.copy()
                self.mlatten = self.atten.copy()
                self.flag = np.full_like(self.resIDs, ISGOOD, dtype=int)
                self.ml_isgood_score = np.full_like(self.resIDs, np.nan, dtype=float)
                self.ml_isbad_score = np.full_like(self.resIDs, np.nan, dtype=float)
                self.phases = np.full_like(self.resIDs, 0, dtype=float)
                self.iqRatios = np.full_like(self.resIDs, 1, dtype=float)
        except IndexError:
            raise ValueError('Unknown number of columns')

        self.freq[np.isnan(self.freq)] = self.mlfreq[np.isnan(self.freq)]
        self.freq[np.isnan(self.freq)] = self.wsfreq[np.isnan(self.freq)]
        self.atten[np.isnan(self.atten)] = self.mlatten[np.isnan(self.atten)]

        self.flag = self.flag.astype(int)
        self.mlfreq[self.flag & ISBAD] = self.wsfreq[self.flag & ISBAD]
        self.ml_isgood_score[self.flag & ISBAD] = 0
        self.ml_isbad_score[self.flag & ISBAD] = 1
        self._vet()


class PhaseFIRCoeffFile:

    def __init__(self, file):
        self.file=file
        self.coeffs=None
        self._load()

    def _load(self):
        # grab FIR coeff from file
        if os.path.splitext(self.file)[1] == ".npz":
            q
            # res_ids = npz['res_ids']
            self.coeffs = npz['filters']
            # out = np.zeros((len(self.resIDs), filters.shape[1]))
            # for index, resID in enumerate(self.resIDs):
            #     if resID not in resIDs:
            #         raise ValueError("Filter coefficients missing resID {}".format(resID))
            #     location = (resIDs == resID)
            #     if location.sum() > 1:
            #         raise ValueError("Filter coefficients contain more than one reference to resID {}".format(resID))
            #     firCoeffs[index, :] = filters[location, :]
        else:
            self.coeffs = np.loadtxt(self.file)


import matplotlib.pyplot as plt
from mkidgen3.opfb import opfb_bin_spectrum, opfb_bin_frequencies
from mkidgen3.util import rx_power
import numpy.typing as nt
import numpy as np


# NB requires a newer version than we have by default in pynq2.7 venv
# def adc_test_plot(adc_data, timerange, fft_range, fft_zoom,  fs=4.096e9, figsize=(13, 10), **mosaic_kw):
#     gs_kw = dict(width_ratios=[1, 1], height_ratios=[1, 1.4])
#     fig, axd = plt.subplot_mosaic([['adc', 'adc'],
#                                    ['fft', 'fftzoom']],
#                                   gridspec_kw=gs_kw, figsize=figsize,
#                                   constrained_layout=True, **mosaic_kw)
#
#
#     adc_timeseries(adc_data, timerange, fs=fs, ax=axd['adc'])
#
#     fft_start, fft_stop = fft_range
#     fft_sl = slice(*fft_range)
#     fft_freqs = np.linspace(-fs/2, fs/2, fft_stop - fft_start)
#     y_fft = np.abs(np.fft.fftshift(np.fft.fft(adc_data[fft_sl])))
#
#     plot_fft(fft_freqs[::2], y_fft[::2] - max(y_fft), ax=axd['fft'])
#     plot_fft(fft_freqs[::2], y_fft[::2] - max(y_fft), ax=axd['fftzoom'], xlim=fft_zoom)
#
#     fig.suptitle('ADC Data')

def plot_single_res_sweep(lo_sweep_freqs, tones, iq_vals, lo_res_freq = None, iq_val_res = None):
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(15, 5))
    plt.rcParams.update({'font.size': 12})

    ax1.plot(lo_sweep_freqs + tones[0], 20 * np.log10(np.abs(iq_vals)), linestyle=" ", marker=".", markersize=10)
    if lo_res_freq is not None:
        ax1.plot(lo_res_freq * 1e6 + tones[0], 20 * np.log10(np.abs(iq_val_res)), linestyle=" ", marker=".", markersize=10,
             label='bias point')
    ax1.set_ylabel('|S21| [dB]')
    ax1.set_xlabel('Frequency [GHz]')
    ax1.set_title('Transmission')
    ax1.legend(loc='center left')

    ax2.plot(iq_vals.real, iq_vals.imag, 'o')
    if lo_res_freq is not None:
        ax2.plot(iq_val_res.real, iq_val_res.imag, 'o', label='bias point')
    ax2.set_xlabel('Real(S21)')
    ax2.set_aspect('equal')
    ax2.set_ylabel('Imag(S21)')
    ax2.set_title('IQ Loop')
    ax2.legend(loc='center left')
    plt.tight_layout()

def plot_phase(phase: nt.NDArray[np.int16], xlim: tuple) -> None:
    """

    Args:
        phase: phase data

    Returns:

    """
    fig, ax = plt.subplots(1, 1, figsize=(15, 10))
    t = 1e3 * np.arange(phase.size) / 2e6
    ax.plot(t, phase)
    ax.set_xlabel('Time (ms)')
    ax.set_ylabel('Phase (radians)')
    ax.set_xlim(xlim);
    ax.set_ylim(-np.pi, np.pi);


def plot_ddc(ddciq: nt.NDArray[np.complex128], phase: nt.NDArray[np.int16], chan_plt: list[int]) -> None:
    """
    Plot post-DDC IQ and phase for given channels
    Args:
        ddciq: complex data from ddciq tap
        phase: phase data
        chan_plt: channels to plot

    Returns:

    """
    fig, axes = plt.subplots(4, len(chan_plt), figsize=(15, 10))
    for j, (i, ax) in enumerate(zip(chan_plt, axes.T)):
        plot_waveforms(ddciq[:100, i], 2e6, mode='iq', ax=ax[0], label='DDC Out' if not j else '')
        plot_waveforms(ddciq[:50, i], 1e6, mode='sw_phase', ax=ax[1], label='Phase/pi (software)' if not j else '')
        plot_waveforms(phase[:50, i] / 2 ** 15, 1e6, mode='phase', ax=ax[2], label='Phase/pi (FPGA)' if not j else '')
        plt.ylim([-1, 1])
        err = np.angle(ddciq[:50, i]) - np.pi * phase[:50, i] / 2 ** 15
        plot_waveforms(err, 1e6, mode='phase', ax=ax[3], label='Phase error (radians)' if not j else '')
    plt.suptitle('Gen3 DSP Pipeline (OPFB Out, DDC, Software phase, FPGA (cordic) phase)');


def plot_comp_sat(
        ol,
        ifboard,
        output_atten=0,
        input_range=(60, 50, 1),
        **kwargs,
):
    input_attens = np.arange(
        input_range[1],
        input_range[0],
        1 if len(input_range) == 2 else input_range[2],
    )[::-1]
    dbms_i, dbms_q = [], []
    adcs_i, adcs_q = [], []
    for input_atten in input_attens:
        ifboard.set_attens(output_attens=output_atten, input_attens=input_atten)
        _, dbm, adc = rx_power(ol.capture.capture_adc(2 ** 19, complex=False))
        dbms_i.append(dbm[0])
        adcs_i.append(adc[0])
        dbms_q.append(dbm[1])
        adcs_q.append(adc[1])
    ifboard.set_attens(output_attens=output_atten, input_attens=input_attens[0])
    dbms_i, dbms_q = np.array(dbms_i), np.array(dbms_q)
    adcs_i, adcs_q = np.array(adcs_i), np.array(adcs_q)

    fig, (ax1, ax3, ax2) = plt.subplots(3, sharex=True, **kwargs)
    ax1.plot(input_attens, dbms_i, label="I")
    ax1.plot(input_attens, dbms_q, label="Q")
    ax1.plot(
        input_attens,
        np.min(dbms_i) - (input_attens - np.max(input_attens)),
        linestyle='--',
        color='gray',
        label="Ideal",
    )
    ax1.set_ylabel("RX Power @ ZU48DR Port (dBm)")

    ax3.plot(input_attens, dbms_i - (np.min(dbms_i) - (input_attens - np.max(input_attens))), label="I")
    ax3.plot(input_attens, dbms_q - (np.min(dbms_i) - (input_attens - np.max(input_attens))), label="Q")
    ax3.plot(
        input_attens,
        np.zeros_like(input_attens),
        linestyle='--',
        color='gray',
        label="Ideal",
    )
    ax3.set_ylabel("Residual (dBm)")

    pg = lambda db: 10 ** (db / 10)

    ax2.plot(input_attens, adcs_q, label='I')
    ax2.plot(input_attens, adcs_i, label='Q')
    ax2.plot(
        input_attens,
        adcs_i[np.argmax(input_attens)] * np.sqrt(pg(np.max(input_attens) - input_attens)),
        label='expected gain')
    ax2.set_ylim(0, 1.1)
    ax2.axhline(1, label="Saturation Point", color="black", zorder=-1, linestyle='-.')
    ax2.set_ylabel("ADC Max Magnitude [0, 1)")
    ax2.set_xlabel("Input Attenuation (dB)")

    ax1.legend()
    ax2.legend()
    ax2.invert_xaxis()
    ax2.set_xlim(np.max(input_attens), np.min(input_attens))
    plt.tight_layout()
    plt.show()

    return dbms_i, dbms_q, adcs_i, adcs_q


def adc_test_plot(adc_data, timerange, fft_range, fft_zoom, db=True, fs=4.096e9, figsize=(16, 8), **mosaic_kw):
    fig = plt.figure(figsize=figsize, constrained_layout=True)
    spec = fig.add_gridspec(2, 2, width_ratios=[1, 1], height_ratios=[1, 1.8])
    adcax = fig.add_subplot(spec[0, :])
    fftax = fig.add_subplot(spec[1, 0])
    fftzoomax = fig.add_subplot(spec[1, 1])

    adc_timeseries(adc_data, timerange, fs=fs, ax=adcax)

    fft_start, fft_stop = fft_range
    fft_sl = slice(*fft_range)
    fft_freqs = np.linspace(-fs / 2, fs / 2, fft_stop - fft_start)
    y_fft = np.abs(np.fft.fftshift(np.fft.fft(adc_data[fft_sl])))

    if db:
        y_fft = 20 * np.log10(y_fft)

    plot_fft(fft_freqs[::2]*1e-6, y_fft[::2] - max(y_fft), ax=fftax)
    plot_fft(fft_freqs[::2]*1e-6, y_fft[::2] - max(y_fft), ax=fftzoomax, xlim=fft_zoom)

    fig.suptitle('ADC Data')


def adc_timeseries(data, timerange=(None, None), fs=4.096e9, ax=None, **kwargs):
    """
    data: np array of captured adc data
    start: first sample # in plot
    stop: last sample # in plot
    fs = ADC sample rate [Hz]

    Returns:
        Plot of ADC timeseries.
    """

    if ax is None:
        plt.figure(figsize=(10, 5))
    else:
        plt.sca(ax)

    n = data.shape[0]  # total samples
    tvec = np.linspace(0, n / fs, n) * 1e9  # time vector [nano seconds]
    sl = slice(*timerange)  # plt slice
    plt.plot(tvec[sl], data.real[sl], color='#34D576', linewidth=6)
    plt.plot(tvec[sl], data.real[sl], "o", color='#346B76', linewidth=8)
    plt.grid(True)
    plt.xlabel("time (ns)", position=(0.5, 1))
    plt.ylabel("signal (V)", position=(0, 0.5))

    plt.gca().set_xlim(None if timerange[0] is None else tvec[timerange[0]],
                       None if timerange[1] is None else tvec[timerange[1]])
    plt.title('Time Series')


def plot_fft(f, y, db=True, xlim=(-2048, 2048), ylim=None, ax=None):
    if ax is not None:
        plt.sca(ax)
    plt.plot(f, y, color='#346B76', linewidth=3)
    plt.grid(True)
    if ylim is not None:
        plt.ylim(*ylim)
    if xlim is not None:
        plt.xlim(*xlim)
    plt.xlabel("Frequency (MHz)", position=(0.5, 0.5))
    plt.ylabel("power (linear)", position=(1, 0.5))
    if db:
        plt.ylabel("power (dB)", position=(1, 0.5))
    plt.title('FFT')


def plot_adc_fft(data, fs=4.096e9, db=True, fft_points=2 ** 14, xlim=None, ylim=None, ax=None):
    fft_freqs = np.linspace(-2.048e9, 2.048e9, fft_points)
    fft_data = np.abs(np.fft.fftshift(np.fft.fft(data[:fft_points])))
    if db:
        fft_data = 20 * np.log10(fft_data)
    fft_data = fft_data - max(fft_data)
    if ax is not None:
        plt.sca(ax)
    plt.plot(fft_freqs, fft_data, color='#346B76', linewidth=3)
    plt.grid(True)
    if ylim is not None:
        plt.ylim(*ylim)
    if xlim is not None:
        plt.xlim(*xlim)
    plt.xlabel("Frequency (Hz)", position=(0.5, 0.5))
    plt.ylabel("power (linear)", position=(1, 0.5))
    if db:
        plt.ylabel("power (dB)", position=(1, 0.5))
    plt.title('Spectrum')


def plot_opfb_bins(data, bins, fine_fft_shift=True, fft_shift=True, left_snip=0, ol=True):
    """
    Inputs:
    - data: Raw data out of OPFB. Should be in the form N x 4096 where N is the number of samples from a single bin.
    - bins: list of OPFB bins 0 to 4095 (note bin 0 is far left in the +/- 2 GHz spectrum).
    - fine_fft_shift: boolean. To apply an fft shift to the fine fft spectrum of each bin or not. (You should.)
    - fft_shift: boolean. To apply an fft shift to the entire OPFB output spectrum or not. (You should.)
    - ol: To plot the bins as overlapping or discard the overlap region."""

    if fft_shift:
        data = np.fft.fftshift(data, axes=1)
    bin_freqs = opfb_bin_frequencies(bins, data.shape[0])
    spectra = opfb_bin_spectrum(data, bins)

    plt.figure(figsize=(16, 6))
    sl = slice(data.shape[0] // 3, -data.shape[0] // 3) if not ol else slice(0, -1)
    plt.plot(bin_freqs[sl] * 1e-6, spectra[sl])
    plt.xlabel("Frequency (MHz)", position=(0.5, 0.5))
    plt.ylabel("power (dB)", position=(1, 0.5))
    return bin_freqs[sl], spectra[sl]


def find_opfb_tones(data):
    """
    Plot the absolute value of the average of the real signal in each OPFB bin. Also report the max peak.
    This quickly gives the location of the peak tone in the OPFB
    Inputs:
    - data: Raw data out of OPFB. Should be in the form N x 4096 where N is the number of samples from a single bin.
    """
    plt.plot(np.abs(data).mean(0), linewidth=3)
    plt.xlabel("OPFB Bin (Raw SSR FFT Output Order)")
    plt.ylabel("|Averge Value of Real Signal|")
    print(f"peak in bin: {np.argmax(abs(data.real.mean(0)))}")


def plot_waveforms(x, sample_rate=2e6, mode=None, ax=None, label=None, xlabel='t (ms)', **pltargs):
    """
    Plots IQ and Phase data from OPFB and DDC Bins.
    Args:
        x: input data as phase or iq values
        sample_rate: samplpe rate in Hz
        mode:
            - "sw_phase": computes and plots phase from IQ values
            - "phase": plots phase
            - "iq": Plots I and Q values as two overlaid timetraces
        ax: plot ax
        label: plot label
        xlabel: xaxis label
        **pltargs: plotargs

    Returns:

    """
    if ax is not None:
        plt.sca(ax)
    t = np.arange(x.size) / sample_rate * 1e3
    if mode == 'sw_phase':
        plt.plot(t, np.angle(x) / np.pi, **pltargs)
        plt.ylim(-1.1, 1.1)
    elif mode == 'phase':
        plt.plot(t, x, **pltargs)
        plt.ylim(-1.1, 1.1)
    elif mode == 'iq':
        plt.plot(t, x.real, **pltargs)
        plt.plot(t, x.imag, **pltargs)
    else:
        raise NotImplementedError(f'mode {mode} is unkown. Valid modes are: "sw_phase", "phase", or "iq"')
    if label:
        plt.ylabel(label)
    if xlabel:
        plt.xlabel(xlabel)


def plot_res_chan(riq, channel, **kwargs):
    reschan_fft = 20 * np.log10(np.abs(np.fft.fftshift(np.fft.fft(riq[:, channel]))))
    reschan_norm = reschan_fft - max(reschan_fft)
    f_ax = np.linspace(-1.024, 1.024, riq.shape[0])
    plt.plot(f_ax, reschan_norm, **kwargs)



from pynq.interrupt import _InterruptController, Interrupt
from pynq import PL
import threading
import queue
import asyncio
from logging import getLogger
import time
from collections import defaultdict


log = getLogger(__name__)


def _background_uio_loop(loop: asyncio.AbstractEventLoop) -> None:
    asyncio.set_event_loop(loop)
    _InterruptController.get_controller('axi_intc_0')
    loop.run_forever()


def create_uio_loop_thread() -> (asyncio.AbstractEventLoop, threading.Thread):
    eventloop = asyncio.new_event_loop()

    _InterruptController._controllers.clear()
    _InterruptController._uio_devices.clear()
    thread = threading.Thread(target=_background_uio_loop, args=(eventloop,), daemon=True,
                              name='UIO_Asyncio loop thread')
    thread.start()
    return eventloop, thread


def _interrupt_name(name):
    try:
        name = name._interrupts['interrupt']['fullpath']
    except (AttributeError, KeyError):
        if not isinstance(name, str) and name in PL.interrupt_pins:
            raise ValueError("name must either be an interrupt pin name "
                             "or have _interrupts['interrupt']['fullpath']")
    return name


class ThreadedPLInterruptManager:
    """
    A thread-safe way of getting interrupt events with pynq. pynq's asyncio interrupt approach requires that all
    await interrupt.wait() and the UIO asyncio reader exist in the same thread (and more, the asyncio.Event objects
    need to have been created in that thread as well.

    This manager provides a static method that is safe to use from any thread that returns interrupt event queue
    (of limited depth!) and a threading.Event that may be waitied on, the queue is unique the interrupt and the event
    to each call of get_monitor(interrupt_name).

    When get_monitor is first called a manager instance will be created that deletes all pynq UioControllers
    and recreates them in a new internal thread that will be using to await interrupt events.
    """
    _loop = None
    _thread = None
    _futures = []

    _queues = {}
    _events = defaultdict(set)
    _event_by_id = defaultdict(threading.Event)
    _aio_eventrepr_by_name = {}
    _interrupts = {}

    @staticmethod
    def get_manager():
        if ThreadedPLInterruptManager._thread is None:
            loop, thread = create_uio_loop_thread()
            ThreadedPLInterruptManager._loop = loop
            ThreadedPLInterruptManager._thread = thread
            time.sleep(1)
        return ThreadedPLInterruptManager


    @staticmethod
    def get_monitor(name, maxq=1, id=None):
        """

        Args:
            name: an object supported by _interrupt_name(name), basically an HLS core or a pynq interrupt path.
            maxq: how many interrupt events to store in the queue before it fills up and older events are pushed out
                ignored on all but the initial call for a given name. The queue is common to all calls.
            id: an optional hashable to be used as an id for the event (i.e. get the same event on subsequent calls)

        Returns: a queue.Queue, threading.Event tuple. Event will be set each time the asyncio event fires.

        """
        m = ThreadedPLInterruptManager.get_manager()  # starts up UIO monitoring loop thread first time
        name = _interrupt_name(name)
        e = threading.Event() if id is None else m._event_by_id[f"{name}{id}"]
        m._events[name].add(e)
        try:
            q = m._queues[name]
            log.debug(f"Ignoring maxq as have existing q")
            return q, e
        except KeyError:
            pass

        m._queues[name] = q = queue.Queue(maxsize=maxq)

        log.info(f"Starting new interrupt monitoring corot for {name} @ {time.perf_counter()}")
        fut = asyncio.run_coroutine_threadsafe(m._interrupt_monitor(name), m._loop)
        m._futures.append(fut)

        time.sleep(.038*2)  # have seen it take at least 38 ms until the interrupt is waiting

        return q, e

    @staticmethod
    def get_status(name):
        """Return a dictionary of event state info for an interrupt"""
        m = ThreadedPLInterruptManager.get_manager()
        name = _interrupt_name(name)
        try:
            i = ThreadedPLInterruptManager._interrupts[name]
        except KeyError:
            i = Interrupt(name)
        intc_mmio = i.parent().mmio
        gi = intc_mmio.read(0x1C)
        ie = bool(intc_mmio.read(0x08) & (1 << i.number))
        iv = bool(intc_mmio.read(0x00) & (1 << i.number))

        return {'axic': {'enabled': ie, 'set': iv, 'global_enable': gi},
                'pynq_asyncio_event': m._aio_eventrepr_by_name.get(name, 'None'),  # None set unset
                'events': {repr(e): e.is_set() for e in m._events[name]}}

    @staticmethod
    def remove_monitor(event):
        """
        Stop monitoring interrupt for event, removing the last monitor will also stop the queue and disable
        the interrupt
        """
        m = ThreadedPLInterruptManager.get_manager()  # starts up UIO monitoring loop thread first time
        name = None
        for k,v in m._events.items():
            try:
                v.remove(event)
                name = k
            except KeyError:
                pass
        if not name:
            return
        for k, _ in filter(lambda kv: kv[1] == event, m._event_by_id.items()):
            m._event_by_id.pop(k)
        if not m._events[name]:
            ThreadedPLInterruptManager._queues.pop(name)  # no more events, kill the queue

    @staticmethod
    async def _interrupt_monitor(name):
        i = ThreadedPLInterruptManager._interrupts[name] = Interrupt(name)
        ThreadedPLInterruptManager._aio_eventrepr_by_name[name] = repr(i.event)
        while True:
            t = time.perf_counter()
            log.info(f"Waiting on interrupt {name}:{i.event} @ {t}. \n"
                     f"  Present watchers: {ThreadedPLInterruptManager._events[name]}")
            ThreadedPLInterruptManager._aio_eventrepr_by_name[name] = repr(i.event)
            await i.wait()
            ThreadedPLInterruptManager._aio_eventrepr_by_name[name] = repr(i.event)
            i.event.clear()
            t = time.perf_counter()
            log.info(f"Interrupt {name} @ {t}")
            for e in ThreadedPLInterruptManager._events[name]:
                e.set()
            try:
                q = ThreadedPLInterruptManager._queues[name]
            except KeyError:
                break
            try:
                q.put_nowait(t)  # will raise full and kill the loop at max queue size
            except queue.Full:
                pass



from pynq import DefaultIP


class IQGen(DefaultIP):
    bindto = ['mazinlab:mkidgen3:iq_gen:0.5']

    def __init__(self, description):
        super().__init__(description=description)

    def generate(self, n=2**20-1):
        """ Tell the block to generate n multiples of 4096 of samples. """
        self.register_map.max = n
        self.register_map.run = True

    def stop(self):
        self.register_map.run=False
